from timemachines.skaters.sk.skinclusion import using_sktime


if using_sktime:
    from timemachines.skaters.sk.skwrappers import sk_theta_hourly_iskater
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE
    from timemachines.skatertools.batch.batchskater import batch_skater_factory

    def sk_theta(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None, emp_mass=0.0,
                 deseasonalize=False):
        return batch_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass,
                                    iskater=sk_theta_hourly_iskater,
                                    iskater_kwargs={'deseasonalize':deseasonalize},
                                    min_e=0, n_warm=20)

    SK_THETA_SKATERS = [sk_theta]

else:
    SK_THETA_SKATERS = []

if __name__=='__main__':
    from timemachines.skatertools.evaluation.evaluators import hospital_mean_square_error
    err1 = hospital_mean_square_error(f=sk_theta, k=3)



import pandas as pd

from timemachines.skaters.orbt.orbitinclusion import using_orbit

if using_orbit:
    from orbit.models.lgt import LGTMAP, LGTAggregated, LGTFull

    def orbit_lgt_iskater(y, k, a, t, seasonality):
        """
           Run orbit over a fixed history and extract predictions
        """
        # It seems that orbit needs a two column dataframe of an index column and
        # measurement column in order to work. It also seems to need enough
        # space in the dataframe to insert the values it predicts.
        df = pd.DataFrame(y).reset_index().rename(columns={0: "response_col", 'index': "date_col"})
        for i in range(k):
            df2 = pd.DataFrame([[len(df), 0]], columns=["date_col", "response_col"])
            df = df.append(df2, ignore_index=True)

        # "test/train" split
        test_size = k
        train_df = df[:-test_size]
        test_df = df[-test_size:]

        # run the model
        if isinstance(seasonality, int) or seasonality == None:
            lgt = LGTMAP(
                response_col="response_col",
                date_col="date_col",
                seasonality=seasonality
            )
        else:
            print("The seasonality given was not an int or None type. Try again...")
            return
        lgt.fit(df=train_df)

        # get the prediction(s)
        predicted_df = lgt.predict(df=test_df)

        # list containing predictions 1 through k steps out
        x = list(predicted_df.prediction)
        # standard deviaiton list (this is technically 2 std)
        x_std = [a / 2 for a in list(predicted_df.prediction_95 - predicted_df.prediction_5)]

        return x, x_std


from typing import List, Tuple, Any
from timemachines.skatertools.utilities.conventions import wrap
from timemachines.skatertools.utilities.suppression import no_stdout_stderr
from timemachines.skaters.bats.batsinclusion import using_bats

if using_bats:
    from tbats import TBATS

    def bats_iskater_factory(y: [[float]], k: int, a: List = None, t: List = None, e=None, freq: str = None, n_max=1000,
                             use_box_cox=None, box_cox_bounds=(0, 1),
                             use_trend=None, use_damped_trend=None,
                             seasonal_periods=None, use_arma_errors=True):
        if a:
            assert len(a) == len(y) + k

        if isinstance(y[0], float):
            y = [wrap(yj) for yj in y]

        estimator = TBATS(use_box_cox=use_box_cox, box_cox_bounds=box_cox_bounds,
                          use_trend=use_trend, use_damped_trend=use_damped_trend,
                          seasonal_periods=seasonal_periods, use_arma_errors=use_arma_errors)
        with no_stdout_stderr():
            fitted_model = estimator.fit(y)
        x = list(fitted_model.forecast(steps=k))
        x_std = [1.0 for _ in x]
        return x, x_std


if __name__=='__main__':
    assert using_bats, 'pip install tbats'
    import numpy as np
    import time
    st = time.time()
    y = list(np.cumsum(np.random.randn(400)))
    x, x_std = bats_iskater_factory(y=y,k=5)
    print(x)
    print(time.time()-st)



from timemachines.skaters.proph.prophiskaterfactory import using_prophet
if using_prophet:
    from timemachines.skaters.proph.prophiskaterfactory import prophet_iskater_factory
    from timemachines.skatertools.utilities.conventions import wrap
    import matplotlib.pyplot as plt
    import pandas as pd
    import numpy as np
    import random
    import math
    import os
    from pprint import pprint

    PLOTS_PATH = os.path.dirname(os.path.realpath(__file__)).replace('timemachines/timemachines','timemachines/gallery')


    #  Just some scripts that try to investigate some curious behaviour of prophet, and performance
    #  comparison against dead-simple averaging of recent values
    #  TODO: Move this to timemachines-offlinetesting


    def is_opinonated(y, forecast:pd.DataFrame,k:int, n_recent:int, multiple:float)->bool:
        """ Check if the forecast is far from any recent values, and thus "opinionated"

        :param y:          data used to fit
        :param forecast:   dataframe produced by prophet fitting
        :param m:          fitted facebook prophet model
        :param k:          number of steps ahead
        :return:
        """
        if isinstance(y[0], float):
            y = [wrap(yj) for yj in y]
        y0 = [yj[0] for yj in y]

        for j in range(1,k+1):
           j_std = np.nanstd(np.diff(y0[-k-50:-k],j))
           recent_ys = y0[-(k+n_recent):-(k+1)]
           upper = np.max(recent_ys)+multiple*j_std*math.sqrt(j)+0.1
           lower = np.min(recent_ys)-multiple*j_std*math.sqrt(j)-0.1
           j_x = forecast['yhat'].values[-(1+k-j)]
           if j_x>upper or j_x<lower:
               deviation = abs(j_x-upper)
               print(deviation)
               return True

        return False


    def next_opinionated_forecast(n_train,k,n_recent, multiple,name=None):
        while True:
            try:
                from microprediction import MicroReader
                n_obs = 0
                while n_obs<1000:
                    mr = MicroReader()
                    if name is None:
                        names = mr.get_stream_names()
                        random_name = random.choice(names)
                    else:
                        random_name = name
                    lag_values, lag_times = mr.get_lagged_values_and_times(name=random_name,count=2000)
                    y = list(reversed(lag_values))
                    t = list(reversed(lag_times))
                    n_obs = len(y)
            except ImportError:
                from timemachines.skatertools.data import hospital
                y = hospital()
                t = [15*60*i for i in range(len(y))]
                name = 'hospital'

            for i in [100*j for j in range(10)]:
                print('Looking at ' + random_name + ' ' + str(i) + '/1000')
                if len(y)>i+2*k+n_train:
                    y_fit = y[i:i + n_train]
                    t_fit = t[i:i + n_train]
                    y_hats, _, forecast, m = prophet_iskater_factory(y=y_fit, k=k, t=t_fit)
                    if is_opinonated(y=y_fit, forecast=forecast,k=k, n_recent=n_recent, multiple=multiple):
                        y_3avg = np.mean(y[i + n_train-3:i + n_train])  # avg of last three points
                        y_true_mean = np.mean( y[i+n_train+k-1:i+n_train+k+1] ) # avg of 3 future points
                        error = (y_hats[-1]-y_true_mean)/abs(0.01+y_3avg)
                        avg_error = (y_3avg-y_true_mean)/abs(0.01+y_3avg)
                        return forecast, m, random_name, error, avg_error, y[i+n_train:i+n_train+k]
            print(random_name+' is okay.')

        return None, None, None


    def plot_next_optionated_forecast(k, n_train, n_recent, multiple, name=None):
        forecast, m, name, err, lv_err, y_next = next_opinionated_forecast(k=k, n_train=n_train,name=name,n_recent=n_recent, multiple=multiple)

        if forecast is not None:
            m.plot(forecast)
            plt.title(name.replace('.json', ''))
            plt.xlabel('Error = ' + str(round(err,3)) + ', Last value err =' + str(round(lv_err,3)))
            plt.plot(forecast['ds'].values[-k:],y_next,'r+')
            plt.pause(5)
            plt.tight_layout()
            rel_error = int(10000*abs((0.01+err)/(0.01+lv_err)))
            jpg_path = PLOTS_PATH+os.path.sep+'n_train='+str(n_train)+'_k_'+str(k)+'_multiple_'+str(round(multiple,1))
            try:
                os.mkdir(jpg_path)
            except FileExistsError:
                pass
            jpg_file = jpg_path+os.path.sep + 'rel='+str(rel_error)+'_name='+name.replace('.json', '.jpg')
            plt.savefig(jpg_file)
            plt.close()
            print(jpg_file)

        return err, lv_err, jpg_path


if __name__=='__main__':
    from momentum import var_init, var_update
    import json
    lv = var_init()
    pr = var_init()
    params = {'multiple':1.5,'n_recent':5,'n_train':500,'k':20}
    while True:
        err, lv_err, jpg_path = plot_next_optionated_forecast(**params)
        report = params.copy()
        report['simple_average'] = var_update(lv,lv_err)
        report['prophet'] = var_update(pr,err)
        pprint(report)
        report_file = jpg_path+'_summary.json'
        with open(report_file,'wt') as f:
            json.dump(report,f)



from timemachines.skaters.rvr.rvrconstant import RIVER_CONSTANT_SKATERS
from timemachines.skaters.rvr.rvrensembles import RIVER_ENSEMBLE_SKATERS
from timemachines.skaters.rvr.rvrhypocratic import RIVER_HYPOCRATIC_SKATERS

RIVER_SKATERS = RIVER_CONSTANT_SKATERS + RIVER_HYPOCRATIC_SKATERS + RIVER_ENSEMBLE_SKATERS


from timemachines.skatertools.utilities.conventions import wrap, split_exogenous


# Let's get on thing straight. An observance is *not* a collection of hermits. Not here anyway.
# Rather, it is a simple cache of
#    - The most recent exogenous variables y[1:]
#    - A queue of k-step ahead "a" variables that are waiting to slide back to the present
# Recall that in the skater convention, "a" variables are supplied k-steps in advance so that
# k-step ahead prediction is possible. But this means we have to cache them until they become
# contemporaneous with y[1:]. Only then can we create an aligned time series of regressors and
# regressand.


def observance(y:[float], o:dict, k:int, a:[float]=None):
    """
    This marshals the k-step ahead vector a and the contemporaneous y[1:] and
    returns a combined vector of all exogenous variables.

    It tracks a list of x and corresponding y, by putting a's in a FIFO queue and
    by caching the previous value of y[1:]

    :param o:  state
    :param k:  Number of steps ahead that a is provided
    :param y:
    :param a:
    :returns:  x_t:[float] vector combining y[1:] with previously supplied a's
    """
    yw = wrap(y)
    aw = wrap(a)

    if not o:
        o = {'a':[ None for _ in range(k) ],
             'z':None,  # Stores the previous value of y[1:]
             'x':list(),
             'y':list()
             }

    y_t, z = split_exogenous(yw)

    # Get the contemporaneous variables from last observation
    if z:
        z_t = o.get('z')  # The previously revealed exogenous variables
        o['z'] = z        # Store for next time
    else:
        z = None
        z_t = None

    # Determine the known in advance variable pertaining to the present
    if aw:
        a_t = o['a'].pop(0)  # The known in advance variable pertaining to this time step
        o['a'].append(aw)  # Put the k-ahead received a value(s) on the queue
    else:
        a = None
        a_t = None

    # Combine into exogenous variables ... but only if both arrived
    if aw and z:
        x_t = z_t + a_t if (z_t and a_t) else None
    elif aw and not z:
        x_t = a_t if a_t else None
    elif (not aw) and z:
        x_t = z_t if z_t else None
    elif (not aw) and not z:
        x_t = None

    if (not z) and (not aw):
        o['y'].append([y_t])  # Special case, no need to wait
    else:
        if x_t:
            o['x'].append(x_t)
            o['y'].append([y_t])
        assert len(o['x']) == len(o['y']), "post-condition"
    return x_t, o




from timemachines.skaters.pmd.pmdinclusion import using_pmd
if using_pmd:

    from timemachines.skaters.pmd.pmdskaters import pmd_univariate
    from timemachines.skatertools.composition.residualshypocratic import quickly_moving_hypocratic_residual_factory
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE


    def pmd_univariate_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        return quickly_moving_hypocratic_residual_factory(f=pmd_univariate, y=y,s=s,k=k,a=a,t=t,e=e)


    def pmd_exogenous_hypocratic(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        return quickly_moving_hypocratic_residual_factory(f=pmd_univariate, y=y,s=s,k=k,a=a,t=t,e=e)


    PMD_SKATERS_COMPOSED = [ pmd_univariate_hypocratic, pmd_exogenous_hypocratic ]
else:
    PMD_SKATERS_COMPOSED = []






from timemachines.skatertools.tuning import optimal_r_for_stream
from timemachines.skaters.proph.prophskaterssingular import fbprophet_univariate_r2
from humpday.optimizers.dlibcube import dlib_default_cube
from microprediction import MicroReader
from timemachines.skaters.proph.prophparams import PROPHET_META, prophet_params
import random
from pprint import pprint


# Optimal meta-parameters for prophet skaters


def fbprophet_univariate_best_r()->float:
    """
      :return:  Provides best r for a randomly chosen data stream
                Takes about 12 hrs to run
    """
    mr = MicroReader()
    names = mr.get_stream_names()
    okay = False
    while not okay:
        name = random.choice(names)
        n_obs = len(mr.get_lagged_values(name=name))
        okay = n_obs > PROPHET_META['n_warm']+50 and '~' not in name
    url = 'https://www.microprediction.org/stream_dashboard.html?stream='+name.replace('.json','')
    print('We will find the best fbprophet hyper-parameters for '+url)
    print('There are ' + str(n_obs) + ' observations in the series.')
    print("Prophet will be fit for most of them, after a burn_in, and for many different hyper-params. Don't hold your breathe.")

    best_r, best_value, info = optimal_r_for_stream(f=fbprophet_univariate_r2,name=name,k=10,optimizer=dlib_default_cube,
                                                    n_burn=PROPHET_META['n_warm']+20,n_trials=50,n_dim=2)
    pprint(info)
    params = prophet_params(r=best_r,dim=2)
    pprint(params)


from timemachines.skaters.sk.skinclusion import using_sktime
from timemachines.skaters.pmd.pmdinclusion import using_pmd
from timemachines.skatertools.composition.conjugation import exp_conjugation_factory

if using_sktime and using_pmd:
    from timemachines.skaters.sk.skwrappers import sk_autoets_iskater
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE
    from timemachines.skatertools.batch.batchskater import batch_skater_factory

    def sk_ae_factory_for_positive_y(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None, emp_mass=0.0,
                      auto=True, trend=None,damped=False):
        return batch_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass,
                                    iskater=sk_autoets_iskater,
                                    iskater_kwargs={'auto':auto,'trend':trend,'damped':damped},
                                    min_e=0, n_warm=20)

    def sk_ae_factory(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None, emp_mass=0.0,
                      auto=True, trend=None,damped=False):
        """
           conjugation with exp/log
        """
        f_kwargs = {'emp_mass':emp_mass,'auto':auto, 'trend':trend, 'damped':damped}
        return exp_conjugation_factory(y=y,s=s,k=k, a=a, t=t, e=e, r=r, f=sk_ae_factory_for_positive_y, f_kwargs=f_kwargs)


    def sk_ae(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None,
                         emp_mass=0.0):
        return sk_ae_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass)


    def sk_ae_add(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None,
              emp_mass=0.0):
        return sk_ae_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass, trend='add', damped=False)


    def sk_ae_mul(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None,
                  emp_mass=0.0):
        # Will fail if y<0
        return sk_ae_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass, trend='mul', damped=False)


    def sk_ae_add_damped(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None,
                  emp_mass=0.0):
        return sk_ae_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass, trend='add', damped=True)


    def sk_ae_mul_damped(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None,
                         emp_mass=0.0):
        return sk_ae_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass, trend='mul', damped=True)


    SK_AE_SKATERS = [sk_ae, sk_ae_add, sk_ae_add_damped]

else:
    SK_AE_SKATERS = []

if __name__=='__main__':
    from timemachines.skatertools.evaluation.evaluators import hospital_mean_square_error_with_sporadic_fit
    for f in SK_AE_SKATERS:
        err1 = hospital_mean_square_error_with_sporadic_fit(f=f, k=3, fit_frequency=70)




from timemachines.skaters.orbt.orbitinclusion import using_orbit

if using_orbit:
    from timemachines.skaters.orbt.orbitwrappers import orbit_lgt_iskater
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE
    from timemachines.skatertools.batch.batchskater import batch_skater_factory

    def orbit_lgt_skater_factory(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None,
                 emp_mass=0.0,
                 seasonality=None):
        return batch_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=r, emp_mass=emp_mass,
                                    iskater=orbit_lgt_iskater,
                                    iskater_kwargs={'seasonality': seasonality},
                                    min_e=0, n_warm=20)

    def orbit_lgt_12(y,s,k,a=None, t=None,e=None):
        return orbit_lgt_skater_factory(y=y, s=s, k=k, a=a,t=t,e=e, seasonality=12)


    def orbit_lgt_24(y,s,k,a=None, t=None,e=None):
        return orbit_lgt_skater_factory(y, s, k, a=a,t=t,e=e, seasonality=24)




import numpy as np


def brownian_with_noise(n:int)->[float]:
    ys_ = np.cumsum(np.random.randn(n))
    ys = [y + np.random.randn() for y in ys_]
    return list(ys)


def brownian_with_exogenous(n:int)->[(float,)]:
    dys1__ = np.random.randn(n+2)
    dys1_ = np.array([ 0.1*a+0.3*b+0.2*c for a,b,c in zip( dys1__[2:], dys1__[1:],dys1__) ])
    dys2_ = list(0.5*np.array(dys1_[2:]))+list(np.random.randn(2))
    ys1 = np.cumsum(dys1_) + 0.025*np.random.randn(n)
    ys2 = np.cumsum(dys2_+0.1*dys1_)+0.15*ys1 + 0.015*np.random.randn(n)
    return list(zip(ys1,ys2))





import warnings

from timemachines.skatertools.tuning.hyper import using_humpday

try:
    from microprediction import MicroReader
    using_microprediction = True
except ImportError:
    warnings.warn('microprediction is not installed so empirical fit cannot occur. pip install microprediction')
    using_microprediction = False


if using_microprediction and using_humpday:

    from timemachines.skatertools.tuning.hyper import optimal_r

    def optimal_r_for_stream(f, name:str, k: int, evaluator = None, optimizer = None,
                                n_trials = None, n_dim = None,  n_burn:int = None, test_objective_first = True)->(
        float, float, dict):   # best_r, best_val, info
        """  Find the best hyper-parameters for a univariate skater using live from www.microprediction.org
        :param f:
        :param name:     Choose from https://www.microprediction.org/browse_streams.html but add '.json' to the end
        :param k:
        :param evaluator:
        :param optimizer:
        :param n_trials:
        :param n_dim:
        :param n_burn:
        :param test_objective_first:
        :return: best_r, best_value, info
        """
        mr = MicroReader()
        lagged_values, lagged_times = mr.get_lagged_values_and_times(name=name)
        t = list(reversed(lagged_times))
        y = list(reversed(lagged_values))
        return optimal_r(f=f,y=y,k=k, a=None,t=t,e=None,evaluator=evaluator,optimizer=optimizer,n_trials=n_trials,
                          n_dim=n_dim, n_burn=n_burn, test_objective_first=test_objective_first)




import numpy as np
from timemachines.skatertools.comparison.eloformulas import elo_update
import time
from pprint import pprint

SLOW_SKATER_KEYWORDS = ['divine','fbprophet','arma','tsa_p3','bats_damped','tsa_aggressive','tsa_balanced','tsa_precision']

try:
    from microprediction import MicroReader
    have_micro = True
except ImportError:
    have_micro = False

if have_micro:
    from timemachines.skatertools.data.live import random_regular_data as DEFAULT_DATA_SOURCE
else:
    DEFAULT_DATA_SOURCE = None

SKATER_F_FACTOR = 1000  # The scale factor for ratings. In chess this is set to 400.
# If the matchup is considered more fluky than a single game of chess, a higher value might make sense.
SKATER_K_FACTOR = 200  # The Elo update factor (maximum rating gain)


def skater_elo_multi_update(elo: dict, k, evaluator=None, n_burn=400, tol=0.01, initial_elo=1600, data_source=None):
    """ Create or update elo ratings by running several algos at once and using Elo update n-1 times

              elo - Dictionary containing the 'state' (i.e. elo ratings and game counts)
              k   - Number of steps to look ahead
              tol - Error ratio that results in a tie being declared
              data_provider - A function taking n_obs and returning y, t
    """
    # Lazy import because networked skaters shouldn't be initialized too early
    from timemachines.skatertools.utilities.locations import pypi_from_name
    from timemachines.skaters.allskaters import SKATERS, skater_from_name  # Only those with no hyper-params

    if data_source is None:
        data_source = DEFAULT_DATA_SOURCE
        assert data_source is not None, "If microprediction is not installed you must supply a different data_source function that returns y, t "

    if not elo:
       _init_elo(elo=elo,SKATERS=SKATERS,initial_elo=initial_elo)
    else:
        # New fields ... one-off fix
        if not 'pypi' in elo:
            elo['pypi'] = [ pypi_from_name(nm) for nm in elo['name']]
        if not 'seconds' in elo:
            elo['seconds'] = [-1 for nm in elo['name']]
        elo['pypi'] = [pypi_from_name(nm) for nm in elo['name']]

        # Check for newcomers
        elo = _newcomers(elo=elo,SKATERS=SKATERS,initial_elo=initial_elo)
        new_names = [f.__name__ for f in SKATERS if f.__name__ not in elo['name']]

    elo, evaluator = _set_evaluator(elo)

    # Choose several skaters whose running
    total_seconds = 0
    chosen = list()
    n_skaters = len(elo['name'])
    candidates = np.random.choice(list(range(n_skaters)), size=10, replace=False)
    for c in candidates:
        snds = elo['seconds'][c]
        if snds<0:
            snds = 45
        if (snds+total_seconds<60) or ((len(chosen)>5) and (np.random.rand()<(100/(2+snds)) )):
            total_seconds += snds
            chosen.append(c)
        if total_seconds>60:
            break

    fs = list()
    chosen_and_imported = list()
    for c in chosen:
        sn = elo['name'][c]
        try:
            f = skater_from_name(sn)
            assert f is not None, 'probably just not used anymore'
            elo['active'][c] = True
            fs.append(f)
            chosen_and_imported.append(c)
        except Exception as e:
            elo['active'][c] = False

    y, t = data_source(n_obs=n_burn + 50)
    scores = list()
    ran_okay = list()
    ran_okay_names = list()
    failed_names = list()
    print('  imported successfully ...')
    pprint(chosen_and_imported)
    for c, f in zip(chosen_and_imported, fs):
        import traceback
        try:
            st = time.time()
            score = evaluator(f=f, y=y, k=k, a=None, t=t, e_fit=15, e_nofit=-1, n_test=50, fit_frequency=100)
            elo['seconds'][c] = time.time() - st
            elo['traceback'][c] = 'passing'
            scores.append(score)
            ran_okay.append(c)
            ran_okay_names.append(f.__name__)
        except Exception as e:
            elo['traceback'][c] = traceback.format_exc()
            elo['seconds'][c] = st - time.time()  # Time taken to fail, as -ve number
            failed_names.append(f.__name__)

    if any(failed_names):
        print('  failing ...')
        pprint(failed_names)

    print('  leaderboard ...')
    leaderboard = sorted([(s,c,n) for s,c,n in zip(scores,ran_okay, ran_okay_names)])
    pprint(leaderboard)
    if len(leaderboard)>=2:
        for j, (winner_score,winner, winner_name) in enumerate(leaderboard[:-1]):
            loser = leaderboard[j+1][1]
            loser_score = leaderboard[j+1][0]
            small = tol * (abs(winner_score) + abs(loser_score))  # Ties
            min_games = min(elo['count'][winner], elo['count'][loser])
            K = SKATER_K_FACTOR / 2.0 if min_games > 25 else SKATER_K_FACTOR  # The Elo update scaling parameter
            winner_prior_elo, loser_prior_elo = elo['rating'][winner], elo['rating'][loser]
            points = 1 if winner_score < loser_score - small else 0.5
            elo['rating'][winner], elo['rating'][loser] = elo_update(winner_prior_elo, loser_prior_elo, points, k=K, f=SKATER_F_FACTOR)
            elo['count'][winner] += 1
            elo['count'][loser] += 1
    return elo


def _init_elo(elo, SKATERS, initial_elo):
    from timemachines.skatertools.utilities.locations import pypi_from_name
    # Initialize game counts and Elo ratings
    elo['name'] = [f.__name__ for f in SKATERS]
    elo['count'] = [0 for _ in SKATERS]
    elo['rating'] = [initial_elo for _ in SKATERS]
    elo['traceback'] = ['not yet run' for _ in SKATERS]
    elo['active'] = [True for _ in SKATERS]
    elo['pypi'] = [pypi_from_name(nm) for nm in elo['name']]
    return elo

def _newcomers(elo, SKATERS,initial_elo):
    # Check for newcomers
    from timemachines.skatertools.utilities.locations import pypi_from_name
    new_names = [f.__name__ for f in SKATERS if f.__name__ not in elo['name']]
    for new_name in new_names:
        elo['name'].append(new_name)
        elo['count'].append(0)
        elo['rating'].append(initial_elo)
        elo['traceback'].append('not yet run')
        elo['active'].append(True)
        elo['seconds'].append(-1)
        elo['pypi'].append(pypi_from_name(new_name))
    return elo

def _set_evaluator(elo):
    from timemachines.skatertools.evaluation.evaluators import evaluate_mean_squared_error_with_sporadic_fit, \
        evaluator_from_name
    evaluator = evaluate_mean_squared_error_with_sporadic_fit
    if evaluator is None:
        if elo.get('evaluator'):
            eval_name = elo['evaluator']
            try:
                evaluator = evaluator_from_name(eval_name)
            except:
                print('Could not retrieve ' + eval_name + ' so reverting.')
                evaluator = evaluate_mean_squared_error_with_sporadic_fit
        else:
            evaluator = evaluate_mean_squared_error_with_sporadic_fit
        print('Evaluating using ' + str(evaluator.__name__))
    elo['evaluator'] = evaluator.__name__
    return elo, evaluator


def skater_elo_update(elo: dict, k, evaluator=None, n_burn=400, tol=0.01, initial_elo=1600, data_source=None):
    """ Create or update elo ratings by performing a random matchup on univariate live data

          elo - Dictionary containing the 'state' (i.e. elo ratings and game counts)
          k   - Number of steps to look ahead
          tol - Error ratio that results in a tie being declared
          data_provider - A function taking n_obs and returning y, t

        Speed is *not* taken into account, yet.
    """
    # Lazy import because networked skaters shouldn't be initialized too early
    from timemachines.skatertools.utilities.locations import pypi_from_name
    from timemachines.skaters.allskaters import SKATERS, \
        skater_from_name  # Only those with no hyper-params

    if data_source is None:
        data_source = DEFAULT_DATA_SOURCE
        assert data_source is not None, "If microprediction is not installed you must supply a different data_source function that returns y, t "

    if not elo:
       _init_elo(elo=elo,SKATERS=SKATERS,initial_elo=initial_elo)
    else:
        # New fields ... one-off fix
        if not 'pypi' in elo:
            elo['pypi'] = [ pypi_from_name(nm) for nm in elo['name']]
        if not 'seconds' in elo:
            elo['seconds'] = [-1 for nm in elo['name']]
        elo['pypi'] = [pypi_from_name(nm) for nm in elo['name']]

        # Check for newcomers
        elo = _newcomers(elo=elo,SKATERS=SKATERS,initial_elo=initial_elo)
        new_names = [f.__name__ for f in SKATERS if f.__name__ not in elo['name']]

    elo, evaluator = _set_evaluator(elo)

    # Choose two random skaters, but avoid slow ones once
    n_skaters = len(elo['name'])
    i1, i2 = np.random.choice(list(range(n_skaters)), size=2, replace=False)
    skater1, skater2 = elo['name'][i1], elo['name'][i2]
    for _ in range(2):
        if any([ slow in skater1 for slow in SLOW_SKATER_KEYWORDS ]) or any([ slow in skater2 for slow in SLOW_SKATER_KEYWORDS ]):
            i1, i2 = np.random.choice(list(range(n_skaters)), size=2, replace=False)
            skater1, skater2 = elo['name'][i1], elo['name'][i2]
    if any([ slow in skater1 for slow in SLOW_SKATER_KEYWORDS ]) and any([ slow in skater2 for slow in SLOW_SKATER_KEYWORDS ]):
        i1, i2 = np.random.choice(list(range(n_skaters)), size=2, replace=False)
        skater1, skater2 = elo['name'][i1], elo['name'][i2]


    fs = list()
    for i, sn in zip([i1, i2], [skater1, skater2]):
        try:
            f = skater_from_name(sn)
            assert f is not None, 'probably just not used anymore'
            elo['active'][i] = True
            fs.append(f)
        except Exception as e:
            elo['active'][i] = False

    if len(fs) == 2:
        # a pitched paddle battle in a bottle
        print(fs[0].__name__ + ' vs. ' + fs[1].__name__)
        y, t = data_source(n_obs=n_burn + 50)
        scores = list()
        for i, f in zip([i1, i2], fs):
            import traceback
            try:
                st = time.time()
                score = evaluator(f=f, y=y, k=k, a=None, t=t, e_fit=15, e_nofit=-1, n_test=50, fit_frequency=100)
                elo['seconds'][i] = time.time()-st
                elo['traceback'][i] = 'passing'
                scores.append(score)
            except Exception as e:
                elo['traceback'][i] = traceback.format_exc()
                elo['seconds'][i] = st-time.time() # Time taken to fail, as -ve number

        if len(scores) == 2:
            small = tol * (abs(scores[0]) + abs(scores[1]))  # Ties
            points = 1 if scores[0] < scores[1] - small else 0 if scores[1] < scores[0] - small else 0.5
            elo1, elo2 = elo['rating'][i1], elo['rating'][i2]
            min_games = min(elo['count'][i1], elo['count'][i2])
            K = SKATER_K_FACTOR / 2.0 if min_games > 25 else SKATER_K_FACTOR  # The Elo update scaling parameter
            elo['rating'][i1], elo['rating'][i2] = elo_update(elo1, elo2, points, k=K, f=SKATER_F_FACTOR)
            elo['count'][i1] += 1
            elo['count'][i2] += 1
            the_winner = skater1 if points > 0.5 else skater2 if points<0.5 else 'nobody'
            print('     won by '+the_winner)

    return elo



import numpy as np
from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE
from timemachines.skatertools.components.parade import parade
from timemachines.skatertools.utilities.nonemath import nonecast, nonecenter
from timemachines.skatertools.utilities.conventions import wrap


def batch_skater_factory(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None,
                         iskater=None, iskater_kwargs: dict=None, emp_mass:float=0.0, emp_std_mass:float=1.0,
                         min_e = None, n_warm=10):
    """
           An easy way to expose "offline" batch-style forecasters as skaters
           This accumulates data points one at a time, and feed them en-masse to the batch forecaster
           It also tracks empirical errors
       
          :param  iskater      -  function wrapping the batch-style forecasting capability
          :param  min_e        -  threshold value triggering a call of the batch forecaster (not optional!)

                                     min_e = -np.inf    ... always call the batch forecaster after it is warm unless e=-np.inf
                                     min_e = 0          ... call if e > 0

          :param  emp_mass     -  how much to empirical mean (i.e. bias) versus forecast
          :param  emp_std_mass -  how much to weigh empirical std (by default it only uses empirical std, not the std
                                   that is generated by the batch forecaster)

    """
    assert iskater is not None,'You must supply iskater'
    if iskater_kwargs is None:
        iskater_kwargs = {}
    if min_e is None:
        print('You should probably supply min_e when using batch_skater_factory, since I don''t know how fast your forecaster is. Setting to 0')
        min_e=0
    assert 0 <= emp_mass <= 1
    assert 0 <= emp_std_mass <= 1

    y = wrap(y)
    a = wrap(a)

    using_a = a is not None
    using_t = t is not None
    if e is None:
        e = 0

    if not s.get('y'):
        # On the first invocation we initialize, and freeze the pattern of usage
        s = {'p': {},  # parade
             'dim_y': len(y),
             'dim_a': len(a) if a is not None else None,
             'using_a': using_a,
             'using_t': using_t,
             'k': k,
             'y': list(),  # historical y
             'a': list(),  # list of a known k steps in advance
             't': list(),
             'warm': False}  # indicates whether iskater has been called yet
    else:
        # Assert immutability of k, dimensions of y,a
        if s['y']:
            assert len(y) == s['dim_y']
            assert k == s['k']
        if s['a']:
            assert s['using_a'], 'skater was supplied a one time but not another'
            assert len(a) == s['dim_a'], 'skater was supplied different length a from one invocation to next'

    if y is None:
        # A convention for puking the state 
        return None, s, None
    else:
        s['y'].append(y)
        if a is not None:
            s['a'].append(a)
        if t is not None:
            assert isinstance(t, float), 'epoch time please'
            s['t'].append(t)

        # Decide whether to call batch
        warm_enough = len(s['y']) >= n_warm
        enough_a = (a is None) or (len(s['a']) > k)
        caller_wants = e >= min_e
        do_call = caller_wants and enough_a and warm_enough

        if do_call:
            # Contemporaneous y, t, a are supplied to the batch forecast, as this is usually the preference
            # Note that a will be k longer than y and t, since it is required for prediction k-steps ahead
            t_arg = s['t'][k:] if t is not None else None
            a_arg = s['a']
            y_arg = s['y'][k:]
            x, x_std = iskater(y=y_arg, k=k, a=a_arg, t=t_arg, **iskater_kwargs)
            s['warm'] = True  # Flag indicating a model has been fit
        else:
            # Default to naive forecast (or a smoother depending on emp_mass, emp_std_mass settings)
            x = [y[0]] * k
            x_std = [1] * k

        # Get running mean prediction errors from the prediction parade
        x_resid, x_resid_std, s['p'] = parade(p=s['p'], x=x, y=y[0])
        x_resid_std = nonecast(x_resid_std, 1.0)

        # Compute center of mass between bias-corrected and uncorrected predictions
        x_corrected = np.array(x_resid) + np.array(x)
        x_center = nonecenter(m=[emp_mass, 1 - emp_mass], x=[x_corrected, x])
        x_std_center = nonecenter(m=[emp_std_mass, 1 - emp_std_mass], x=[x_resid_std, x_std])

        return x_center, x_std_center, s



from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, T_TYPE, E_TYPE
from typing import Any
from timemachines.skatertools.ensembling.ensemblefactory import R_BALANCED, R_AGGRESSIVE, R_PRECISION
from timemachines.skaters.elo.eloensemblefactory import elo_ensemble_factory
from timemachines.skatertools.utilities.internet import connected_to_internet

FASTEST = 1.0
FASTER = 10.0

if connected_to_internet():

    def elo_fastest_univariate_precision_ensemble(y: Y_TYPE, s: dict, k: int = 1, a:A_TYPE = None, t:T_TYPE = None, e:E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_PRECISION, n=5, max_seconds=FASTEST)


    def elo_fastest_univariate_balanced_ensemble(y: Y_TYPE, s: dict, k: int = 1, a:A_TYPE = None, t:T_TYPE = None, e:E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_BALANCED, n=5, max_seconds=FASTEST)


    def elo_fastest_univariate_aggressive_ensemble(y: Y_TYPE, s: dict, k: int = 1, a:A_TYPE = None, t:T_TYPE = None, e:E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_AGGRESSIVE, n=5, max_seconds=FASTEST)


    def elo_faster_univariate_precision_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_PRECISION, n=5, max_seconds=FASTER)


    def elo_faster_univariate_balanced_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None,
                                                 e: E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_BALANCED, n=5, max_seconds=FASTER)


    def elo_faster_univariate_aggressive_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_AGGRESSIVE, n=5, max_seconds=FASTER)


    def elo_fastest_residual_precision_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None,
                                                  e: E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_PRECISION, n=5, max_seconds=FASTEST, category='residual')


    def elo_fastest_residual_balanced_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None,
                                                 e: E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_BALANCED, n=5, max_seconds=FASTEST, category='residual')


    def elo_fastest_residual_aggressive_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None,
                                                   e: E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_AGGRESSIVE, n=5, max_seconds=FASTEST, category='residual')


    def elo_faster_residual_precision_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None,
                                                 e: E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_PRECISION, n=5, max_seconds=FASTER, category='residual')


    def elo_faster_residual_balanced_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None,
                                                e: E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_BALANCED, n=5, max_seconds=FASTER, category='residual')


    def elo_faster_residual_aggressive_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None,
                                                  e: E_TYPE = None):
        return elo_ensemble_factory(y=y, s=s, k=k, a=a, t=t, e=e, r=R_AGGRESSIVE, n=5, max_seconds=FASTER, category='residual')


    ELO_UNIVARIATE_SKATERS = [ elo_fastest_univariate_aggressive_ensemble, elo_fastest_univariate_balanced_ensemble,
                             elo_fastest_univariate_precision_ensemble,
                             elo_faster_univariate_aggressive_ensemble, elo_faster_univariate_precision_ensemble,
                             elo_faster_univariate_balanced_ensemble]

    ELO_RESIDUAL_SKATERS = [elo_fastest_residual_aggressive_ensemble, elo_fastest_residual_balanced_ensemble,
                              elo_fastest_residual_precision_ensemble,
                              elo_faster_residual_aggressive_ensemble, elo_faster_residual_precision_ensemble,
                              elo_faster_residual_balanced_ensemble]

    ELO_ENSEMBLE_SKATERS = ELO_UNIVARIATE_SKATERS + ELO_RESIDUAL_SKATERS

else:
    ELO_ENSEMBLE_SKATERS = []




from timemachines.skaters.sk.sktheta import SK_THETA_SKATERS
from timemachines.skaters.sk.skautoarima import SK_AA_SKATERS
from timemachines.skaters.sk.skautoets import SK_AE_SKATERS
SK_SKATERS = SK_THETA_SKATERS + SK_AA_SKATERS + SK_AE_SKATERS



from datetime import datetime
from typing import Union, List
import pandas as pd
import pytz
from timemachines.skatertools.utilities.arrivals import approx_dt


EPOCH  = datetime.utcfromtimestamp(0)                     # Unix epoch start time as datetime
RECENT = datetime(2021,1,1)
RECENT_SECONDS = round((RECENT - EPOCH).total_seconds())  # Epoch seconds for a more recent date

FloatOrList = Union[float,List[float]]
DatetimeOrList = Union[datetime,List[datetime]]


def epoch_to_naive_datetime(t:FloatOrList)->DatetimeOrList:
    """ Naive conversion """
    if isinstance(t,float):
        return epoch_to_naive_datetime([t])[0]
    else:
        return [datetime.utcfromtimestamp(tj) for tj in t]


def naive_datetime_to_epoch(d:DatetimeOrList)->FloatOrList:
    if isinstance(d,float):
        return naive_datetime_to_epoch([d])[0]
    else:
        return [ dj.replace(tzinfo=pytz.UTC).timestamp() for dj in d]


def is_valid_freq(freq):
    try:
        dr = pd.date_range(start=EPOCH,periods=3,freq=freq)
        return True
    except ValueError:
        return False


def infer_freq_from_epoch(t:[float])->str:
    """ Infer a frequency string
    :param t: [ float ] epoch times
    :return:
    """
    # https://github.com/pandas-dev/pandas/blob/master/pandas/tseries/frequencies.py
    dt = epoch_to_naive_datetime(t)
    dti = pd.DatetimeIndex(dt)
    pd_freq = pd.infer_freq(dti)
    if pd_freq is None:
        dt = approx_dt(t)
        if dt is not None:
            return pd.infer_freq(epoch_to_naive_datetime([j * dt for j in range(20)]))
        else:
            return None
    else:
        return pd_freq



from timemachines.skaters.smdk.smdkinclusion import using_simdkalman
if using_simdkalman:

    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
    from timemachines.skatertools.components.parade import parade
    from timemachines.skatertools.utilities.nonemath import nonecast
    import numpy as np
    import random
    import math
    from timemachines.skatertools.ensembling.precisionweightedskater import precision_weighted_skater
    from simdkalman.primitives import ddot, ddot_t_right, dinv

    # Pronounced "sim-d-k ARMA factory", not "simd-karma-factory"

    # Not ready for prime time by any means


    def random_p(max_p: int):
        return random.choice(list(range(1, max_p + 1)))


    def random_q(max_q: int):
        return random.choice(list(range(1, max_q + 1)))


    def phi_decay(j):
        return 0.5 ** (j + 1)


    def theta_decay(j):
        return 0.5 ** (j + 1)


    def random_phi(p: int):
        return [(np.random.rand() - 0.5) * phi_decay(j) for j in range(p)]


    def random_theta(q: int):
        return [(np.random.rand() - 0.5) * theta_decay(j) for j in range(q)]


    def smdk_arma_factory(y: Y_TYPE, n_agents: int, max_p: int, max_q:int, s, k: int, a: A_TYPE = None, t: T_TYPE = None,
                          e: E_TYPE = None, r: R_TYPE = 0.75, min_vintage=50):
        """

              max_p - Maximum AR order
              max_q - Maximum MA order
              r     - Controls how much the more accurate models are weighted

        """
        n_states = max(max_p, max_q+1)
        n_obs = 1  # May generalize later
        assert n_states >= 2
        y0 = wrap(y)[0]
        if not s.get('n_states'):
            s = _arma_initial_state(n_states=n_states, n_agents=n_agents, k=k, x0=y0)
        else:
            assert n_agents == s['n_agents']
            assert n_states == s['n_states']

        if y0 is None:
            return None, s, None
        else:
            # Mutation step - flag those needed system updates
            # Changes s['phi'], s['theta'], s['noise'], s['sigma'] and flags with s['stale']
            if s['n_measurements'] > 10:
                s = _arma_evolution(s,min_vintage=min_vintage)

            # Update system equations if params have mutated
            s = _arma_matrix_update(s, k)

            # Get time step, or default to 1 second
            if t is None:
                dt = 1
            else:
                if s.get('prev_t') is None:
                    s['prev_t'] = -1
                dt = t - s['prev_t']
                s['prev_t'] = t

            # Kalman updates
            prior_mean = ddot(s['A'], s['m'])  # A * m
            prior_cov = ddot(s['A'], ddot_t_right(s['P'], s['A'])) + dt * s['Q']  # A * P * A.t + Q
            measurement = np.array([[[y0]]])
            posterior_mean, posterior_cov, K, ll = _update(prior_mean=prior_mean, prior_covariance=prior_cov,
                                                           observation_model=s['H'], observation_noise=s['R'],
                                                           measurement=measurement,
                                                           log_likelihood=True)
            s['m'] = posterior_mean
            s['P'] = posterior_cov

            # Compute k-step ahead predictions for each agent
            agent_xs = [[np.nan for _ in range(k)] for _ in range(n_agents)]
            agent_stds = [[np.nan for _ in range(k)] for _ in range(n_agents)]
            for j in range(k):
                if j == 0:
                    j_posterior_mean = posterior_mean
                else:
                    j_posterior_mean = ddot(s['powers_of_A'][j - 1, :, :, :], posterior_mean)
                j_y_hat = ddot(s['H'], j_posterior_mean)
                for ndx in range(n_agents):
                    agent_xs[ndx][j] = j_y_hat[ndx, 0, 0]

            # Update agent prediction parades and get their empirical standard errors
            for ndx in range(n_agents):
                _discard_bias, agent_std_j, s['parades'][ndx] = parade(p=s['parades'][ndx], x=agent_xs[ndx], y=y0)
                agent_stds[ndx] = nonecast(agent_std_j, fill_value=1.0)

            # Create the exogenous vector that the precision weighted skater expects.
            # (i.e. y_for_pws[1:] has agent predictions interlaced with their empirical means)
            y_for_pws = [y0]
            s['fitness'] = []
            for agent_x, agent_std in zip(agent_xs, agent_stds):
                y_for_pws.append(agent_x[-1])
                y_for_pws.append(agent_std[-1])
                s['fitness'].append(1./(1e-6+agent_std[-1]**2))

            # Call the precision weighted skater
            x, x_std, s['s_pks'] = precision_weighted_skater(y=y_for_pws, s=s['s_pws'], k=k, a=a, t=t, e=e, r=r)
            x_std_fallback = nonecast(x_std, fill_value=1.0)

            s['n_measurements'] += 1
            if s['n_measurements'] < 10:
                # Cold ... just puke naive forecasts
                return [y0] * k, [1.0] * k, s
            else:
                return x, x_std_fallback, s



    def _arma_initial_state(n_states, n_agents, k, x0:float):
        """
        :param n_states:    number of latest states is equal to max(p,q+1)
        :param n_agents:    number of independent ARMA models computed at once
        :param k:           number of steps ahead to forecast
        :param x0:          initial value for all latent lag states
        :return:
        """
        n_obs = 1 # dimension of observation (fixed at 1 for now)
        s = {'n_states': n_states,
             'n_agents': n_agents,
             'n_measurements': 0,
             'm': np.zeros((n_agents, n_states,1)),
             'P': np.zeros((n_agents, n_states, n_states)),
             'H': np.zeros((n_agents, n_obs, n_states)),
             'R': np.zeros((n_agents, n_obs, n_obs)),
             'Q': np.zeros((n_agents, n_states, n_states)),
             'A': np.zeros((n_agents, n_states, n_states)),
             'fitness': [1. for _ in range(n_agents)],
             'prev_t': None,
             's_pws': {}  # State for the precision weighted skater
             }
        ps = [random_p(n_states) for _ in range(n_agents)]
        qs = [random_q(n_states - 1) for _ in range(n_agents)]
        s['phi'] = [random_phi(p) for p in ps]
        s['theta'] = [random_theta(q) for q in qs]
        s['parades'] = [{} for _ in range(n_agents)]  # Track empirical errors individually (somewhat inefficient)
        s['stale'] = [True for _ in range(n_agents)]
        s['r_var'] = [np.random.exponential() ** 4 for _ in range(n_agents)]
        s['q_var'] = [np.random.exponential() ** 4 for _ in range(n_agents)]
        s['powers_of_A'] = np.ndarray((k, n_agents, n_states, n_states))
        s['vintage'] = [ 0 for _ in range(n_agents )]
        # Initialize random states
        for ndx in range(n_agents):
            s['m'][ndx, :, 0] = [x0 for _ in range(n_states)]
            s['P'][ndx, :, :] = (np.random.exponential() ** 4) * np.eye(n_states)
        return s

    def _arma_evolution(s, min_vintage):
        """
            Modify the ARMA population based on fitness
        """
        s['vintage'] = [v + 1 for v in s['vintage']]
        experienced = [ j for j, v in enumerate(s['vintage']) if v>=min_vintage ]
        if any(experienced):
           experienced_fitness = [f for f,v in zip(s['fitness'],s['vintage']) if v >= min_vintage]
           good_fitness = np.percentile(experienced_fitness,q=80)
           poor_fitness = np.percentile(experienced_fitness, q=20)
           good = [ j for j in experienced if s['fitness'][j]>=good_fitness ]
           poor = [ j for j in experienced if s['fitness'][j] <= poor_fitness ]
           n_change = int(math.ceil(s['n_agents']/100))
           for _ in range(n_change):
               a,b,c = list(np.random.choice(good,size=3,replace=False))
               d = np.random.choice(poor)
               s['stale'][d]=True
               s['vintage'][d]=0
               evol_type = random.choice(['ar','ma','r_var','q_var'])
               if evol_type=='ma':
                   s['theta'][d] = [ theta_a+(theta_b-theta_c) for theta_a, theta_b, theta_c in zip(s['theta'][a], s['theta'][b]+[0],s['theta'][c]+[0]  )]
               elif evol_type=='ar':
                   s['phi'][d] = [phi_a + (phi_b - phi_c) for phi_a, phi_b, phi_c in zip(s['phi'][a], s['phi'][b]+[0], s['phi'][c]+[0])]
               elif evol_type=='r_var':
                   s['r_var'][d] = math.exp( math.log(s['r_var'][a]) + ( math.log(s['r_var'][b]) - math.log(s['r_var'][c])) )
               elif evol_type == 'q_var':
                   s['q_var'][d] = math.exp(math.log(s['q_var'][a]) + (math.log(s['q_var'][b]) - math.log(s['q_var'][c])))

        return s


    def _arma_matrix_update(s, k):
        n_states = s['n_states']
        n_obs = 1
        for ndx, (stl, ph, tht, r_var, q_var) in enumerate(
                zip(s['stale'], s['phi'], s['theta'], s['r_var'], s['q_var'])):
            p = len(ph)
            q = len(tht)
            if stl:
                A_ = np.zeros((n_states, n_states))
                A_[0, :p] = ph
                for j in range(n_states - 1):
                    A_[j + 1, j] = 1.
                s['A'][ndx, :, :] = A_
                H_ = np.ndarray((1,n_states))
                H_[0,:] = [1] + tht + [0 for _ in range(n_states-len(tht)-1)]
                s['H'][ndx, :, :] = H_
                Q_ = np.zeros((n_states, n_states))
                Q_[0, 0] = q_var
                s['Q'][ndx, :, :] = Q_
                R_ = np.ones((n_obs, n_obs))
                R_[0, 0] = r_var
                s['R'][ndx, :, :] = R_
                s['powers_of_A'][0, ndx, :, :] = A_
                for j in range(1, k):
                    s['powers_of_A'][j, ndx, :, :] = ddot(s['powers_of_A'][j - 1, ndx, :, :], A_)
        return s

    def _update(prior_mean, prior_covariance, observation_model, observation_noise, measurement, log_likelihood=False):

        # y - H * mp
        v = measurement - ddot(observation_model, prior_mean)

        # H * Pp * H.t + R
        S = ddot(observation_model, ddot_t_right(prior_covariance, observation_model)) + observation_noise
        invS = dinv(S)

        # Kalman gain: Pp * H.t * invS
        K = ddot(ddot_t_right(prior_covariance, observation_model), invS)

        # K * v + mp
        posterior_mean = ddot(K, v) + prior_mean

        # Pp - K * H * Pp
        posterior_covariance = prior_covariance - ddot(K, ddot(observation_model, prior_covariance))

        # inv-chi2 test var
        # outlier_test = np.sum(v * ddot(invS, v), axis=0)
        if log_likelihood:
            l = np.ravel(ddot(v.transpose((0,2,1)), ddot(invS, v)))
            l += np.log(np.linalg.det(S))
            l *= -0.5
            return posterior_mean, posterior_covariance, K, l
        else:
            return posterior_mean, posterior_covariance, K

        return posterior_mean, posterior_covariance


K_LEADERBOARD_TYPES = ['univariate','residual']
from getjson import getjson
from timemachines.skaters.localskaters import local_skater_from_name


# Utilities for grabbing suggested models from the Elo ratings


FIBONACCI = [ 1, 2, 3, 5, 8, 13, 21, 34]  # k-s used in the ratings


def closest(lst, k):
    return lst[min(range(len(lst)), key=lambda i: abs(lst[i] - k))]


def closest_fib(k):
    return closest(FIBONACCI,k)


def ratings_url(k:int, category='univariate'):
    url_template = 'https://raw.githubusercontent.com/microprediction/timeseries-elo-ratings/main/ratings/CATEGORY-k_KKK.json'
    kkk = str(k).zfill(3)
    return url_template.replace('KKK',kkk).replace('CATEGORY', category)


def get_ratings(k:int, category='univariate'):
    url = ratings_url(k=k,category=category)
    return getjson(url=url)


def top_rated(k:int, category='univariate', max_seconds=10, min_count=10, require_passing:bool=False,
              ignore_elo=True):
    """
    :param k:                number of steps ahead to forecast
    :param category:        'univariate' or 'residual'
    :param max_seconds:      for about 50 separate forecasts
    :param min_count:        minimum number of games played to establish the Elo rating
    :param require_passing:  filter out if the tests are failing
    :param ignore_elo        filter out skaters with elo in name, to avoid infinite recursion
    :return:    [ (rating, name, package_url) ]
    """
    k_closest = closest_fib(k)
    rd = get_ratings(k=k_closest,category=category)
    rd_zip = zip( rd['name'], rd['count'], rd['rating'],rd['traceback'],rd['seconds'], rd['pypi'] )
    return sorted([(rtng,nm,pypi) for nm,cnt,rtng,trcbck,scnds,pypi in rd_zip if
                   ((scnds>=0) and (scnds<=max_seconds)) and
                   (trcbck=='passing' or (not require_passing)) and
                   (cnt>=min_count) and
                   ((not 'elo' in nm) or not ignore_elo)
                   ], reverse=True )


def top_rated_names(k:int, n=5, category='univariate', max_seconds=10, min_count=10, require_passing:bool=False, ignore_elo=False)->[str]:
    rcm = top_rated(k=k, category=category, max_seconds=max_seconds, min_count=min_count, require_passing=require_passing,ignore_elo=ignore_elo)
    return [ r[1] for r in rcm[:n]]


def top_rated_models(k:int, n=5, category='univariate', max_seconds=10, min_count=10, require_passing:bool=False, ignore_elo=False):
    """
       Try to instantiate top skaters, moving down the list until we have at least n

       :returns  [ skaters ]
    """
    rcm = top_rated(k=k, category=category, max_seconds=max_seconds, min_count=min_count, require_passing=require_passing, ignore_elo=ignore_elo)
    loaded = list()

    for r in rcm:
        try:
            skater = local_skater_from_name(r[1])
            if skater is not None:
                loaded.append(skater)
                if len(loaded)==n:
                    break
        except ImportError:
            print('Cannot instantiate recommendation ' +r[1]+'. For install instructions see the package '+r[2])

    return loaded


if __name__=='__main__':
    my_skaters = top_rated_models(k=3, n=5)
    print(my_skaters)






from timemachines.skaters.rvr.rvrinclusion import using_river
from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
if using_river:
    from river import linear_model, optim, preprocessing, time_series
    from timemachines.skatertools.components.parade import parade
    from timemachines.skatertools.utilities.nonemath import nonecast, nonecenter

    def rvr_sarimax_factory(y :Y_TYPE, s, k:int, a:A_TYPE =None,
                            t:T_TYPE =None, e:E_TYPE =None, r:R_TYPE=None,
                            p:int=0, d:int=0, q:int=0, m=1, sp:int=0, sq:int=0,
                            intercept_init=110, optim_sgd=0.01, intercept_lr=0.3 ):
        y = wrap(y)
        a = wrap(a)

        if not s.get('k'):
            s['k']=k
            s['n'] = len(y)
            s['p']={}  # parade
            s['model']=None
        else:
            assert len(y) == s['n']
            assert k == s['k']

        if y is None:
            return None, s, None
        else:
            model = s.get('model')
            if model is None:
                model = time_series.SNARIMAX(p=p, d=d, q=q, m=m, sp=sp, sq=sq,
                                              regressor=(
                                                      preprocessing.StandardScaler() |
                                                      linear_model.LinearRegression(
                                                          intercept_init=intercept_init,
                                                          optimizer=optim.SGD(optim_sgd),
                                                          intercept_lr=intercept_lr
                                                      )
                                              ))
            x = model.forecast(horizon=1)
            _we_ignore_bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y[0])
            x_std_fallback = nonecast(x_std, fill_value=1.0)
            model = model.learn_one(x=None,y=y[0])
            s['model'] = model
            return x, x_std_fallback, s






# Implements the old Matlab tic/toc which is still my favourite interface
# That style makes is somewhat invasive - though the odds of s['chronometer'] being used for
# another purpose seem slim

import time
from pprint import pprint


def tick(s):
    if not s.get('chronometer'):
        s['chronometer'] = {'tick':time.time()}
    else:
        s['chronometer'].update({'tick':time.time()})


def tocks(s):
    """ Returns a copy of reported timings """
    return dict([(k,v) for k,v in s['chronometer'].items() if k not in ['tick']])


def tock(s,label:str='invocation',reset=True, accumulate=True):
    assert label not in ['tick','tock']
    assert s.get('chronometer') and s.get('chronometer').get('tick'),'Must call tick before tock'
    elapsed = time.time()-s['chronometer']['tick']
    if accumulate and s['chronometer'].get(label):
        elapsed += s['chronometer'][label]
    s['chronometer'][label] = elapsed
    if reset:
        s['chronometer']['tick']=time.time()
    return s


def lap(s,label:str):
    return tock(s=s,label=label,reset=False)


if __name__=='__main__':
    s = dict()
    tick(s)
    time.sleep(1)
    tock(s)
    pprint(s)
    time.sleep(1)
    tock(s,'fit')
    pprint(s)


PYPI = {'tsa':'statsmodels',
        'fbprophet':'prophet',
        'pmd':'pmdarima',
        'rvr':'river',
        'nprophet':'neuralprophet',
        'dlm':'pydlm',
        'divine':'divinity',
        'orbit':'orbit-ml',
        'bats':'tbats',
        'ats':'auto_ts',
        'glu':'gluonts',
        'flux':'pyflux',
        'sk':'sktime',
        'smdk':'simdkalman',
        'gk':'greykite',
        'tcn':'keras-tcn',
        'darts':'darts'}


def pypi_from_name(name):
    stem = name.split('_')[0]
    short_name = PYPI.get(stem)
    stub = 'https://pypi.org/project/'
    return stub+short_name if short_name else stub+'timemachines'



if __name__=='__main__':
    from timemachines.skaters.localskaters import LOCAL_SKATERS
    from pprint import pprint
    pprint([(sk.__name__,pypi_from_name(sk.__name__)) for sk in LOCAL_SKATERS])
    print(len(LOCAL_SKATERS))


import sys, traceback
from contextlib import contextmanager, redirect_stderr, redirect_stdout
import os


# This one is recommended for facebook prophet, to try to get STAN to shut-the-hell-up
class no_stdout_stderr(object):
    '''
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all print, even if the print originates in a
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).

    '''
    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close the null files
        for fd in self.null_fds + self.save_fds:
            os.close(fd)



class suppress_output:

    def __init__(self, suppress_stdout=True, suppress_stderr=True):
        self.suppress_stdout = suppress_stdout
        self.suppress_stderr = suppress_stderr
        self._stdout = None
        self._stderr = None

    def __enter__(self):
        devnull = open(os.devnull, "w")
        if self.suppress_stdout:
            self._stdout = sys.stdout
            sys.stdout = devnull
        if self.suppress_stderr:
            self._stderr = sys.stderr
            sys.stderr = devnull

    def __exit__(self, *args):
        if self.suppress_stdout:
            sys.stdout = self._stdout
        if self.suppress_stderr:
            sys.stderr = self._stderr



import numpy as np


def smdk_example(different_transitions=False):
    """ Example multi-dimensional Kalman filter system parameters

    :param different_transitions:
    :return:   A,H,m0,P0,Q,R,y0
    """

    # Process noise
    Q = np.zeros((2 ,3, 3))
    Q[0, :, :] = np.eye(3)
    Q[1, :, :] = np.eye(3)

    # Prior state means
    should_we_transpose = True
    if should_we_transpose:
        m0 = np.zeros((2, 3, 1))
        m0[0, :, :] = np.array([[1.5, 1.5, 1.5]]).transpose()
        m0[1, :, :] = np.array([[1.5, 1.5, 1.5]]).transpose()
    else:
        m0 = np.zeros((2, 1, 3))
        m0[0, :, :] = np.array([[1.5, 1.5, 1.5]])
        m0[1, :, :] = np.array([[1.5, 1.5, 1.5]])

    # Prior covariances
    P0 = np.zeros((2, 3, 3))
    P0[0, :, :] = np.eye(3)
    P0[1, :, :] = np.eye(3)

    # Observations
    y0 = np.zeros((2, 1, 1))
    y0[0, :, :] = np.array([[-1.4]])
    y0[1, :, :] = np.array([[-1.4]])

    # Measurement var
    R = np.zeros((2, 1, 1))
    R[0, :, :] = np.eye(1)
    R[1, :, :] = 0.6 * np.eye(1)

    # Observation equations
    H0 = np.array([[0.8, 0.2, 0]])
    H1 = np.array([[0.8, 0.2, 0]])
    H0.shape = (1, 3)
    H1.shape = (1, 3)
    H = np.zeros((2, 1, 3))
    H[0, :, :] = H0
    H[1, :, :] = H1

    # Single observation
    y0 = np.zeros((2, 1, 1)) # First dim indexes the time-series, second the time-step, third the observation dimension
    y0[0, :, :] = np.array([[-1.4]])
    y0[1, :, :] = np.array([[-1.7]])

    # Transition(s)
    A0 = np.array([[0.5, 0.3, 0.2], [1, 0, 0], [0, 1, 0]])
    if different_transitions:
        A = np.zeros((2, 3, 3))
        A1 = np.array([[0.4, 0.4, 0.2], [1, 0, 0], [0, 1, 0]])
        A[0, :, :] = A0
        A[1, :, :] = A1
    else:
        A = A0

    return A, H, m0, P0, Q, R, y0




from timemachines.skaters.sk.skinclusion import using_sktime
from timemachines.skaters.pmd.pmdinclusion import using_pmd

if using_sktime:
    import numpy as np
    import pandas as pd
    from sktime.forecasting.theta import ThetaForecaster
    from sktime.forecasting.base import ForecastingHorizon
    import datetime
    from typing import List
    from timemachines.skatertools.utilities.suppression import no_stdout_stderr

    def sk_theta_hourly_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls sktime's theta forecaster implementation, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        y0_series = pd.Series(index=pd.PeriodIndex(pd.date_range("2021-01", periods=len(y0s), freq="H")), data=y0s)
        last_t = y0_series.index[-1]
        next_t = last_t.to_timestamp() + datetime.timedelta(hours=1)
        with no_stdout_stderr():
            forecaster = ThetaForecaster(deseasonalize=deseasonalize)
            forecaster.fit(y0_series)
            fh = ForecastingHorizon(pd.PeriodIndex(pd.date_range(next_t, periods=k, freq="H")), is_relative=False )
            x = forecaster.predict(fh)
        x_std = [1.0 for _ in x]
        return x, x_std

    if using_pmd:

        from sktime.forecasting.arima import AutoARIMA
        from sktime.forecasting.ets import AutoETS


        def sk_autoarima_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None,
                                 start_p=2, d=None, start_q=2, max_p=5, max_d=2, max_q=5, start_P=1, D=None, start_Q=1,
                                 max_P=2, max_D=1, max_Q=2, max_order=5, sp=1, seasonal=True, stationary=False,
                                 information_criterion='aic', alpha=0.05, freq='H', test='kpss', seasonal_test='ocsb', stepwise=True,
                                 n_jobs=1, start_params=None, trend=None, method='lbfgs', maxiter=50, offset_test_args=None,
                                 seasonal_test_args=None, suppress_warnings=True, n_fits=10, out_of_sample_size=0, scoring='mse',
                                 scoring_args=None, with_intercept=True, **kwargs):
            """
               TODO: Fix if t is supplied, similar to with prophet
            """
            if a:
                assert len(a) == len(y) + k
            if np.isscalar(y[0]):
                y0s = [yt for yt in y]
            else:
                y0s = [yt[0] for yt in y]
            y0_series = pd.Series(index=pd.PeriodIndex(pd.date_range("2021-01", periods=len(y0s), freq=freq)), data=y0s)
            last_t = y0_series.index[-1]
            next_t = last_t.to_timestamp() + datetime.timedelta(hours=1)
            with no_stdout_stderr():
                forecaster = AutoARIMA(start_p=start_p, d=d, start_q=start_q, max_p=max_p, max_d=max_d, max_q=max_q,
                                       start_P=start_P, D=None, start_Q=start_Q,
                                 max_P=max_P, max_D=max_D, max_Q=max_Q, max_order=max_order, sp=sp, seasonal=seasonal, stationary=stationary,
                                 information_criterion=information_criterion, alpha=alpha, test=test, seasonal_test=seasonal_test, stepwise=stepwise,
                                 n_jobs=n_jobs, start_params=start_params, trend=trend, method=method, maxiter=maxiter, offset_test_args=offset_test_args,
                                 seasonal_test_args=seasonal_test_args, suppress_warnings=suppress_warnings, error_action='warn', trace=False,
                                 n_fits=n_fits, out_of_sample_size=out_of_sample_size, scoring=scoring,
                                 scoring_args=scoring_args, with_intercept=with_intercept)
                # For argument meanings see: https://www.sktime.org/en/latest/api_reference/modules/auto_generated/sktime.forecasting.arima.AutoARIMA.html
                forecaster.fit(y0_series)
                fh = ForecastingHorizon(pd.PeriodIndex(pd.date_range(next_t, periods=k, freq=freq)), is_relative=False)
                x = forecaster.predict(fh)
            x_std = [1.0 for _ in x]
            return x, x_std

        def sk_autoets_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None,error='add',
                                      trend=None, damped_trend=False, seasonal=None, sp=1, initialization_method='estimated',
                                      initial_level=None, initial_trend=None, initial_seasonal=None, bounds=None,
                                      dates=None, freq='H', missing='none', start_params=None, maxiter=1000,
                                      full_output=True, disp=False, callback=None, return_params=False, auto=False,
                                      information_criterion='aic', allow_multiplicative_trend=False, restrict=True,
                                      additive_only=False, ignore_inf_ic=True, n_jobs=1, **kwargs):
            """
               Defaults to hourly
            """
            if a:
                assert len(a) == len(y) + k
            if np.isscalar(y[0]):
                y0s = [yt for yt in y]
            else:
                y0s = [yt[0] for yt in y]
            y0_series = pd.Series(index=pd.PeriodIndex(pd.date_range("2021-01", periods=len(y0s), freq=freq)), data=y0s)
            last_t = y0_series.index[-1]
            next_t = last_t.to_timestamp() + datetime.timedelta(hours=1)
            with no_stdout_stderr():
                forecaster = AutoETS(error=error,
                                      trend=trend, damped_trend=damped_trend, seasonal=seasonal, sp=1, initialization_method=initialization_method,
                                      initial_level=initial_level, initial_trend=initial_trend, initial_seasonal=initial_seasonal, bounds=bounds,
                                      dates=dates, freq=freq, missing=missing, start_params=start_params, maxiter=maxiter,
                                      full_output=full_output, disp=disp, callback=None, return_params=False, auto=auto,
                                      information_criterion=information_criterion, allow_multiplicative_trend=allow_multiplicative_trend,
                                     restrict=restrict,
                                      additive_only=additive_only, ignore_inf_ic=ignore_inf_ic, n_jobs=n_jobs)
                forecaster.fit(y0_series)
                fh = ForecastingHorizon(pd.PeriodIndex(pd.date_range(next_t, periods=k, freq="H")), is_relative=False)
                x = forecaster.predict(fh)
            x_std = [1.0 for _ in x]
            return x, x_std


if __name__=='__main__':
    assert using_sktime, 'pip install sktime'
    assert using_pmd, 'pip install pmdarima'
    import numpy as np
    import time
    st = time.time()
    y = list(np.cumsum(np.random.randn(400)))
    x, x_std = sk_autoarima_iskater(y=y, k=5)
    print(x)
    print(time.time()-st)



from typing import List, Iterator


def reflect(ys:[float], n:int)->[float]:
    """ Lengthen time series by reflection back in time
    :param ys: time series shorter than n
    :param n:  desired length of time series
    :return: time series length n
    """
    assert len(ys)>=3
    xs = list(reversed(ys))[:-1] + list(ys)
    while len(xs)<n:
        xs = xs[:-1] + xs
    return xs[:n]


def dilate(u,scale):
    if isinstance(u,list):
        return [u_*scale for u_ in u]
    else:
        return u*scale




from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE
from timemachines.skaters.bats.batsinclusion import using_bats
if using_bats:
    from timemachines.skaters.bats.batsfactory import bats_factory


    def bats_fast(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_trend=False, use_arma_errors=False, use_box_cox=False)


    def bats_trendy(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None, r:R_TYPE=None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_trend=True, use_arma_errors=False, use_box_cox=False)


    def bats_damped(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_damped_trend=True, use_arma_errors=False, use_box_cox=False)


    def bats_arma(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_trend=False, use_arma_errors=True, use_box_cox=False)


    def bats_arma_bc(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_trend=False, use_arma_errors=True, use_box_cox=True)


    def bats_damped_arma(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_damped_trend=True, use_arma_errors=True, use_box_cox=False)


    def bats_trendy_arma(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_trend=True, use_arma_errors=True, use_box_cox=False)


    def bats_bc(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_trend=False, use_arma_errors=False, use_box_cox=True)


    def bats_trendy_bc(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_trend=True, use_arma_errors=False,
                            use_box_cox=True)

    def bats_damped_bc(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_damped_trend=True, use_arma_errors=False,
                            use_box_cox=True)

    def bats_damped_arma_bc(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_damped_trend=True, use_arma_errors=True, use_box_cox=True)


    def bats_trendy_arma_bc(y: Y_TYPE, s, k: int, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None, r: R_TYPE = None):
        return bats_factory(y=y, s=s, k=k, a=a, t=t, e=e, use_trend=True, use_arma_errors=True, use_box_cox=True)


    BATS_SKATERS = [bats_fast, bats_arma, bats_bc, bats_arma_bc,
                    bats_trendy, bats_trendy_arma, bats_trendy_bc, bats_trendy_arma_bc,
                    bats_damped, bats_damped_arma, bats_damped_bc, bats_damped_arma_bc]
else:
    BATS_SKATERS = []


import requests

CHECKED_INTERNET=False
CONNECTED_TO_INTERNET=None


def connected_to_internet():
    global CHECKED_INTERNET
    global CONNECTED_TO_INTERNET
    if not CHECKED_INTERNET:
        CONNECTED_TO_INTERNET = check_if_connected_to_internet()
    return CONNECTED_TO_INTERNET


def check_if_connected_to_internet(url='http://www.google.com/', timeout=2):
    try:
        _ = requests.head(url, timeout=timeout)
        return True
    except requests.ConnectionError:
        print("No internet connection available.")
    return False



try:
    import simdkalman
    using_simdkalman = True
except ImportError:
    using_simdkalman = False
    using_latest_simdkalman = False


if using_simdkalman:
    # Also checks whether we can use the version supporting multiple systems
    # pip install --upgrade git+https://github.com/oseiskar/simdkalman should fix
    from timemachines.skaters.smdk.smdkexample import smdk_example
    A, H, m0, P0, Q, R, y0 = smdk_example(different_transitions=False)
    from simdkalman import KalmanFilter
    kf = KalmanFilter(state_transition=A, process_noise=Q, observation_model=H,observation_noise=R)
    try:
        A, H, m0, P0, Q, R, y0 = smdk_example(different_transitions=True)
        kf = KalmanFilter(state_transition=A, process_noise=Q, observation_model=H, observation_noise=R)
        using_latest_simdkalman = True
    except AssertionError:
        using_latest_simdkalman = False



from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
from timemachines.skatertools.composition.residualshypocratic import quickly_moving_hypocratic_residual_factory, slowly_moving_hypocratic_residual_factory
from timemachines.skaters.simple.movingaverage import slowly_moving_average, quickly_moving_average
from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory
from typing import Any

# Composition of fast and slow moving averages
# Mostly for offlinetesting, no idea if these make any sense


def thinking_slow_and_fast(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
           Apply defensive quickly moving average to residuals from a slowly moving average
    """
    return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, f=slowly_moving_average)


def thinking_fast_and_slow(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
          Apply defensive slowly moving average to residuals from a quickly moving average
    """
    return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, f=slowly_moving_average)


def thinking_slow_and_slow(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
           Apply defensive slow moving average to residuals from a slowly moving average
    """
    return slowly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, f=slowly_moving_average)


def thinking_fast_and_fast(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
           Apply defensive quickly moving average to residuals from a quickly moving average
    """
    return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, f=quickly_moving_average)


def thinking_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
    """
         Precision weight moving averages
    """
    fs = [ slowly_moving_average, quickly_moving_average, thinking_fast_and_fast, thinking_fast_and_slow,
           thinking_slow_and_fast, thinking_slow_and_slow ]
    return precision_weighted_ensemble_factory(fs=fs,y=y,s=s,k=k,a=a,t=t,e=e,r=0.5)


THINKING_SKATERS = [ thinking_fast_and_fast, thinking_fast_and_slow, thinking_slow_and_slow, thinking_slow_and_fast ]


if __name__=='__main__':
    k = 4
    from timemachines.skatertools.visualization.realplot import hospital_prior_plot_exogenous
    import matplotlib.pyplot as plt
    hospital_prior_plot_exogenous(f=thinking_ensemble,k=k)
    plt.show()
    pass





# Not yet skaters, since they haven't had r set
# Work in progress to expose and tune en masse

from timemachines.skaters.proph.allprophetskaters import PROPHET_R2_SKATERS
from timemachines.skaters.simple.movingaverage import EMA_R1_SKATERS

SKATERS_R3 = []
SKATERS_R2 = PROPHET_R2_SKATERS
SKATERS_R1 = EMA_R1_SKATERS



from timemachines.skaters.drts.dartsarimaskater import DARTS_ARIMA_SKATERS
from timemachines.skaters.drts.dartsautoarimaskater import DARTS_AUTOARIMA_SKATERS
from timemachines.skaters.drts.dartsexpsmoothingskater import DARTS_EXPONENTIALSMOOTHING_SKATERS
from timemachines.skaters.drts.dartsfftskater import DARTS_FFT_SKATERS
from timemachines.skaters.drts.dartsfourthetaskater import DARTS_FOURTHETA_SKATERS
from timemachines.skaters.drts.dartsprophetskater import DARTS_PROPHET_SKATERS
from timemachines.skaters.drts.dartsthetakater import DARTS_THETA_SKATERS

DARTS_SKATERS = DARTS_ARIMA_SKATERS + DARTS_AUTOARIMA_SKATERS + DARTS_EXPONENTIALSMOOTHING_SKATERS + DARTS_FFT_SKATERS + DARTS_FOURTHETA_SKATERS + DARTS_PROPHET_SKATERS + DARTS_THETA_SKATERS


DARTS_SKATERS = DARTS_ARIMA_SKATERS + DARTS_AUTOARIMA_SKATERS + DARTS_EXPONENTIALSMOOTHING_SKATERS + DARTS_FFT_SKATERS + DARTS_FOURTHETA_SKATERS + DARTS_PROPHET_SKATERS


try:
    from microprediction import MicroReader
    using_micro = True
except ImportError:
    using_micro = False

import random

# Live data streams at www.microprediction.com

if using_micro:
    import numpy as np

    def random_stream_name(min_len=500,exclude_str=None, include_str=None):
        """ Randomly selected univariate series
                :return:  y, t, url
                """
        mr = MicroReader()
        names = mr.get_stream_names()
        okay = False
        while not okay:
            name = random.choice(names)
            n_obs = len(mr.get_lagged_values(name=name, count=10000))
            okay = True
            if exclude_str is not None and exclude_str in name:
                okay = False
            if include_str is not None and include_str not in name:
                okay = False
            if n_obs < min_len:
                okay = False
        url = 'https://www.microprediction.org/stream_dashboard.html?stream=' + name.replace('.json', '')
        return name, url

    def random_regular(min_len=500):
        """ Randomly selected univariate series
        :return:  y, t, url
        """
        return random_stream_name(min_len=min_len,exclude_str='~')


    def random_stream_data(min_len=500,exclude_str='~', include_str='electricity'):
        """ Randomly selected univariate series
        :return:  y, t, url
        """
        name, url = random_stream_name(min_len=min_len, exclude_str=exclude_str, include_str=include_str)
        return stream_data(name=name, n_obs=1000)


    def random_residual(min_len=500):
        """ Randomly selected univariate series
        :return:  y, t, url
        """
        return random_stream_name(min_len=min_len, include_str='z1~')


    def stream_data(name:str,n_obs:int):
        """ values and times for a univariate stream """
        mr = MicroReader()
        lagged_values, lagged_times = mr.get_lagged_values_and_times(name=name, count=n_obs)
        y, t = list(reversed(lagged_values)), list(reversed(lagged_times))
        return y, t


    def random_regular_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
        :returns:  y, t
        """
        name, url = random_regular(min_len=n_obs)
        return stream_data(name=name,n_obs=n_obs)


    def random_residual_data(n_obs=500):
        """ Retrieve univariate time series chosen randomly from the collection at www.microprediction.org
        :returns:  y, t
        """
        name, url = random_residual(min_len=n_obs)
        return stream_data(name=name, n_obs=n_obs)


    def random_rescaled_stream_data(n_obs=900,include_str='electricity',exclude_str='~'):
        y, t = random_stream_data(min_len=n_obs, include_str=include_str,exclude_str=exclude_str)
        dy = np.diff(y)
        scale = (max(np.abs(dy) ) -min(np.abs(dy)) ) +0.001
        return [ yt /scale for yt in y[:n_obs] ]


    def random_surrogate_data(f, k=1, n_real=50, n_samples=200, n_warm = 500, n_input=20, verbose=False,
                               include_str='electricity',exclude_str='~'):
        """ Create dataset of skater input (abbreviated to last n_input lags) and output
        :param f:   skater
        :param k:   step ahead
        :param n_real:   number of distinct real time series to use
        :param n_input:  number of lags to use in training
        :param n_warm:   number of data points to use to warm up skater before harvesting predictions
        :return: x_train, y_train, y_true
        """
        n_obs = n_samples + n_warm + 2* k + 10
        n_train = n_real * n_samples
        x_train = np.zeros(shape=(n_train, n_input, 1))
        y_train = np.zeros(shape=(n_train, 1))
        x_stds = [1.0]
        y_true = list()
        if verbose:
            print(np.shape(x_train))
        for ndx_real in range(n_real):
            # Get real data
            y_real = random_rescaled_stream_data(n_obs=n_obs, include_str=include_str, exclude_str=exclude_str)
            if verbose:
                print(y_real[:5])
                print((ndx_real, n_real))
                print(np.mean(x_stds))
            # Warm the model up
            y_real_warm = y_real[:n_warm]
            s = {}
            for y in y_real_warm:
                x, x_std, s = f(y, k=k, s=s)
            # Now create examples of k-step ahead forecasts
            y_real_harvest = y_real[n_warm:n_warm + n_samples]
            for j2, y in enumerate(y_real_harvest):
                x, x_std, s = f(y, k=k, s=s)
                xk = x[k - 1]
                j = ndx_real * n_samples + j2
                yj_true = y_real[n_warm + j2 + k]
                y_true.append(yj_true)
                y_train[j, 0] = xk
                x_stds.append(x_std[-k])
                input_data = y_real[n_warm + j2 - n_input+1:n_warm +j2+1]
                assert input_data[-1]==y
                for l in range(n_input):
                    x_train[j, l, 0] = input_data[l]

        return x_train, y_train, y_true


if __name__=='__main__':
    if using_micro:
        name, url = random_regular(min_len=900)
        print(url)


from timemachines.skaters.rvr.rvrinclusion import using_river

if using_river:
    from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory, R_BALANCED, R_AGGRESSIVE, R_PRECISION
    from timemachines.skaters.rvr.rvrconstant import RIVER_CONSTANT_SKATERS
    from timemachines.skatertools.utilities.conventions import Y_TYPE,A_TYPE,T_TYPE,E_TYPE
    from typing import Any


    def rvr_precision_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
        return precision_weighted_ensemble_factory(fs=RIVER_CONSTANT_SKATERS,y=y,s=s,k=k,a=a,t=t,e=e,r=R_PRECISION)


    def rvr_balanced_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
        return precision_weighted_ensemble_factory(fs=RIVER_CONSTANT_SKATERS,y=y,s=s,k=k,a=a,t=t,e=e,r=R_BALANCED)


    def rvr_aggressive_ensemble(y :Y_TYPE, s:dict, k:int =1, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None)->([float] , Any , Any):
        return precision_weighted_ensemble_factory(fs=RIVER_CONSTANT_SKATERS,y=y,s=s,k=k,a=a,t=t,e=e,r=R_AGGRESSIVE)

    RIVER_ENSEMBLE_SKATERS = [ rvr_aggressive_ensemble, rvr_balanced_ensemble, rvr_aggressive_ensemble ]
else:
    RIVER_ENSEMBLE_SKATERS = []




from timemachines.skaters.dlm.dlmunivariate import using_dlm, autoReg, dlm
if using_dlm:

    class fixedAutoReg(autoReg):
        # A hack to make DLM auto work, the version in PyPI at time of writing throws error trying to append

        def __init__(self,*arg,**kwargs):
            super().__init__(*arg,**kwargs)

        def appendNewData(self, data):
            """ AutoReg append new data automatically with the main time series. Nothing
            needs to be done here.
            """
            return


    class quietDlm(dlm):
        # Same as DLM but without priting

        def __init__(self,*arg,**kwargs):
            super().__init__(*arg,**kwargs)

        # Append new data or features to the dlm
        def append(self, data, component='main'):
            """ Append the new data to the main data or the components (new feature data)

            Args:
                data: the new data
                component: the name of which the new data to be added to.\n
                           'main': the main time series data\n
                           other omponent name: add new feature data to other
                           component.

            """
            # initialize the model to ease the modification
            if not self.initialized:
                self._initialize()

            # if we are adding new data to the time series
            if component == 'main':
                # add the data to the self.data
                self.data.extend(list(data))

                # update the length
                self.n += len(data)
                self.result._appendResult(len(data))

                # update the automatic components as well
                for component in self.builder.automaticComponents:
                    comp = self.builder.automaticComponents[component]
                    comp.appendNewData(data)

            # if we are adding new data to the features of dynamic components
            elif component in self.builder.dynamicComponents:
                comp = self.builder.dynamicComponents[component]
                comp.appendNewData(data)

            else:
                raise NameError('Such dynamic component does not exist.')



from timemachines.skaters.nproph.nprophetiskaterfactory import using_neuralprophet, NeuralProphet

if using_neuralprophet:
    from timemachines.skaters.nproph.nprophetskaterfactory import nprophet_skater_factory
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE
    from typing import List


    def nprophet_p1(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None):
        return nprophet_skater_factory(y=y,s=s,k=k,a=a,t=t,e=e, recursive = False, model_params = {'n_lags':1})


    def nprophet_p2(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None):
        return nprophet_skater_factory(y=y,s=s,k=k,a=a,t=t,e=e, recursive = False, model_params = {'n_lags':2})


    def nprophet_p3(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None):
        return nprophet_skater_factory(y=y,s=s,k=k,a=a,t=t,e=e, recursive = False, model_params = {'n_lags':3})


    def nprophet_p5(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None):
        return nprophet_skater_factory(y=y,s=s,k=k,a=a,t=t,e=e, recursive = False, model_params = {'n_lags':5})


    def nprophet_p8(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                                 t: T_TYPE = None, e: E_TYPE = None):
        return nprophet_skater_factory(y=y,s=s,k=k,a=a,t=t,e=e, recursive = False, model_params = {'n_lags':5})


    NPROPHET_UNIVARIATE_SKATERS = [ nprophet_p1, nprophet_p2, nprophet_p3, nprophet_p5, nprophet_p8 ]
else:
    NPROPHET_UNIVARIATE_SKATERS = []






from timemachines.skaters.dlm.dlmunivariate import using_dlm
if using_dlm:
    from timemachines.skaters.dlm.dlmunivariate import dlm_univariate_r3, dlm_univariate_a, dlm_univariate_b
    from timemachines.skaters.dlm.dlmexogenous import dlm_exogenous_a, dlm_exogenous_b, dlm_exogenous_r3
    from timemachines.skaters.dlm.dlmcomposed import DLM_SKATERS_COMPOSED

    DLM_SKATERS = [ dlm_univariate_a, dlm_univariate_b ] + DLM_SKATERS_COMPOSED
    DLM_R3_SKATERS = [ dlm_univariate_r3, dlm_exogenous_r3 ]
else:
    DLM_SKATERS = []
    DLM_R3_SKATERS = []


from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any
import math


def conjugation_factory(y :Y_TYPE, s:dict, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None, r=None,
                            f=None, h=None, h_inv=None, apply_to_exog=False, h_kwargs=None, f_kwargs=None, check_inversion=False)->([float] , Any , Any):
    """

        h: R -> R
        h_inv: R -> R
    """
    y = wrap(y)
    if h_kwargs is None:
        h_kwargs = {}

    y_transformed = [yi for yi in y]
    y_transformed[0] = h(y[0], **h_kwargs)
    if check_inversion:
        y0_back = h_inv(y_transformed[0], **h_kwargs)
        assert abs(y[0]-y0_back)<0.01, 'h_inv appears not to be an inverse of h'

    if apply_to_exog:
        for k in range(1,len(y)):
            y_transformed[k] = h(y[k], **h_kwargs)

    if f_kwargs is None:
        f_kwargs = {}
    if r is not None:
        x_transformed, x_transformed_std, s = f(y=y_transformed,s=s,k=k, a=a, t=t, e=e, r=r, **f_kwargs)
    else:
        x_transformed, x_transformed_std, s = f(y=y_transformed,s=s,k=k, a=a, t=t, e=e, **f_kwargs)

    x_inv = [ h_inv(xk,**h_kwargs) for xk in x_transformed]
    shrink_factor = 0.1
    try:
        x_upper = [ h_inv(xk+shrink_factor*xk_std, **h_kwargs) for xk, xk_std in zip(x_transformed,x_transformed_std)]
        x_lower = [ h_inv(xk-shrink_factor*xk_std, **h_kwargs) for xk, xk_std in zip(x_transformed, x_transformed_std)]
        x_inv_std = [ abs(xu-xl)/(2*shrink_factor) for xu,xl in zip(x_upper, x_lower)]
    except ValueError:
        # The prediction might not remain in R+
        x_inv_std = [ 1.0 for _ in x_inv ]

    return x_inv, x_inv_std, s


def exp_conjugation_factory(y :Y_TYPE, s:dict, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None, r=None, f=None, f_kwargs=None)->([float] , Any , Any):
    """
       For when models only work on (0,inf)
       e.g. multiplicative errors etc 
    """
    assert f is not None

    if s.get('running_endog_envelope') is None:
        s['running_endog_envelope'] = 10000.

    # Maintain a rough scale that lags behind actual values and changes slowly
    y = wrap(y)
    if (abs(y[0])>2*s['running_endog_envelope']):
        s['running_endog_envelope']=1.1*s['running_endog_envelope']

    # This defines a slowly changing map into R+
    exponent = 0.1/s['running_endog_envelope']

    def h(x):
        return math.exp(exponent*x)

    def h_inv(x):
        return math.log(x)/exponent

    return conjugation_factory(y=y,s=s, k=k, a=a, t=t, e=e, r=r, f=f, h=h, h_inv=h_inv,
                               apply_to_exog=False, check_inversion=True, f_kwargs=f_kwargs)


    
    
    
    




# Local skaters are so named to differentiate them from those that require
# a connection (e.g. to look up Elo ratings and stack on the fly)

# Skaters have no hyper-parameters to set.
# Most of these are univariate and ignore y[1:] and a[:]
# We'll create some that use y[1:] more effectively as we go.

# If you add, please update skatertools.utilities.locations

# Home grown
from timemachines.skaters.simple.movingaverage import EMA_SKATERS
from timemachines.skaters.simple.linear import LINEAR_SKATERS
from timemachines.skaters.simple.thinking import THINKING_SKATERS
from timemachines.skaters.simple.hypocraticensemble import HYPOCRATIC_ENSEMBLE_SKATERS
# Using 3rd party
from timemachines.skaters.divine.divineskaters import DIVINE_SKATERS
from timemachines.skaters.proph.allprophetskaters import PROPHET_SKATERS
from timemachines.skaters.dlm.alldlmskaters import DLM_SKATERS
from timemachines.skaters.pmd.allpmdskaters import PMD_SKATERS
from timemachines.skaters.tsa.alltsaskaters import TSA_SKATERS
from timemachines.skaters.nproph.allnprophetskaters import NPROPHET_SKATERS
from timemachines.skaters.orbt.allorbitskaters import ORBIT_SKATERS
from timemachines.skaters.bats.allbatsskaters import BATS_SKATERS
from timemachines.skaters.rvr.allriverskaters import RIVER_SKATERS
from timemachines.skaters.sk.allskskaters import SK_SKATERS
from timemachines.skaters.gk.allgreykiteskaters import GREYKITE_SKATERS
from timemachines.skaters.smdk.allsmdkskaters import SMDK_SKATERS
from timemachines.skaters.drts.alldartsskaters import DARTS_SKATERS

LOCAL_SKATERS = EMA_SKATERS + PROPHET_SKATERS + DIVINE_SKATERS + DLM_SKATERS + \
                THINKING_SKATERS + PMD_SKATERS + TSA_SKATERS + NPROPHET_SKATERS + \
                HYPOCRATIC_ENSEMBLE_SKATERS + ORBIT_SKATERS + BATS_SKATERS \
                + RIVER_SKATERS + SK_SKATERS + GREYKITE_SKATERS + SMDK_SKATERS \
                + DARTS_SKATERS


LEFT_OUT_FOR_NOW = LINEAR_SKATERS # + ...

# Skaters designed for online use
FAST_LOCAL_SKATERS = EMA_SKATERS + THINKING_SKATERS + HYPOCRATIC_ENSEMBLE_SKATERS + RIVER_SKATERS + SMDK_SKATERS



def local_skater_from_name(name):
    valid = [f for f in LOCAL_SKATERS if f.__name__ == name]
    return valid[0] if len(valid)==1 else None







from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous
from statsmodels.tsa.forecasting.theta import ThetaModel
from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any
from timemachines.skaters.tsa.tsaparams import TSA_META
from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory
from timemachines.skatertools.components.parade import parade
from timemachines.skatertools.utilities.nonemath import nonecast
import warnings
from statsmodels.tools.sm_exceptions import ConvergenceWarning
warnings.simplefilter('ignore', ConvergenceWarning)


def tsa_theta_factory(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None,
                      method=None, deseasonalize=True,
                      use_test=True, difference=False) -> ([float], Any, Any):
    """

    period : int, default None
        The period of the data that is used in the seasonality test and
        adjustment. If None then the period is determined from y's index,
        if available.
    deseasonalize : bool, default True
        A flag indicating whether the deseasonalize the data. If True and
        use_test is True, the data is only deseasonalized if the null of no
        seasonal component is rejected.
    use_test : bool, default True
        A flag indicating whether test the period-th autocorrelation. If this
        test rejects using a size of 10%, then decomposition is used. Set to
        False to skip the test.
    method : {"auto", "additive", "multiplicative"}, default "auto"
        The model used for the seasonal decomposition. "auto" uses a
        multiplicative if y is non-negative and all estimated seasonal
        components are positive. If either of these conditions is False,
        then it uses an additive decomposition.
    difference : bool, default False
        A flag indicating to difference the data before offlinetesting for
        seasonality."""

    y = wrap(y)
    a = wrap(a)

    if not s.get('y'):
        s = {'y': list(),
             'a': list(),
             'k': k,
             'p':{}}
    else:
        # Assert immutability of k, dimensions
        if s['y']:
            assert len(y) == len(s['y'][0])
            assert k == s['k']
        if s['a']:
            assert len(a) == len(s['a'][0])

    if y is None:
        return None, s, None
    else:
        s['y'].append(y)
        if a is not None:
            s['a'].append(a)
        if len(s['y']) > max(2 * k + 5, TSA_META['n_warm']) and (e is not None and e>0):
            y0s = [ y_[0] for y_ in s['y']]
            import numpy as np
            model = ThetaModel(np.array(y0s), method=method, deseasonalize=deseasonalize,
                      use_test=use_test, difference=difference)
            try:
                x = list( model.fit().forecast(steps=k) )
            except Exception as e:
                print(e)
                x = [wrap(y)[0]]*k
        else:
            x = [y[0]] * k

        y0 = wrap(y)[0]
        _we_ignore_bias, x_std, s['p'] = parade(p=s['p'], x=x, y=y0)
        x_std_fallback = nonecast(x_std, fill_value=1.0)
        return x, x_std_fallback, s


def tsa_theta_auto(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_theta_factory(y=y,s=s,k=k,a=a,t=t,e=e, method='auto', deseasonalize=False)


def tsa_theta_additive(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
    return tsa_theta_factory(y=y,s=s,k=k,a=a,t=t,e=e, method='additive', deseasonalize=False)


def tsa_theta_multiplicative(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                           t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return tsa_theta_factory(y=y, s=s, k=k, a=a, t=t, e=e, method='multiplicative', deseasonalize=False)


TSA_THETA_SKATERS = [ tsa_theta_auto, tsa_theta_additive, tsa_theta_multiplicative ]



if __name__=='__main__':
    prior_plot_exogenous(f=tsa_theta_auto,k=1,n=TSA_META['n_warm']+25,n_plot=25)



from timemachines.skaters.simple.movingaverage import balanced_ema_ensemble, aggressive_ema_ensemble, \
    precision_ema_ensemble
from timemachines.skatertools.composition.residualshypocratic import quickly_moving_hypocratic_residual_factory, \
    slowly_moving_hypocratic_residual_factory
from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, E_TYPE, T_TYPE


def quick_balanced_ema_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
    """
           Apply defensive quickly moving average to residuals from an EMA ensemble
    """
    return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, f=balanced_ema_ensemble)


def slow_balanced_ema_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
    """
           Apply defensive quickly moving average to residuals from an EMA ensemble
    """
    return slowly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, f=balanced_ema_ensemble)


def quick_aggressive_ema_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
    """
           Apply defensive quickly moving average to residuals from an EMA ensemble
    """
    return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, f=aggressive_ema_ensemble)


def slow_aggressive_ema_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
    """
           Apply defensive slowly moving average to residuals from an EMA ensemble
    """
    return slowly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, f=aggressive_ema_ensemble)


def quick_precision_ema_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
    """
           Apply defensive quickly moving average to residuals from an EMA ensemble
    """
    return quickly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, f=precision_ema_ensemble)


def slow_precision_ema_ensemble(y: Y_TYPE, s: dict, k: int = 1, a: A_TYPE = None, t: T_TYPE = None, e: E_TYPE = None):
    """
           Apply defensive slowly moving average to residuals from an EMA ensemble
    """
    return slowly_moving_hypocratic_residual_factory(y=y, s=s, k=k, a=a, t=t, e=e, f=precision_ema_ensemble)


HYPOCRATIC_ENSEMBLE_SKATERS = [ quick_balanced_ema_ensemble, slow_balanced_ema_ensemble,
                                quick_aggressive_ema_ensemble, slow_aggressive_ema_ensemble,
                                quick_precision_ema_ensemble, slow_precision_ema_ensemble]



if __name__=='__main__':
    k = 3
    from timemachines.skatertools.visualization.realplot import hospital_prior_plot_exogenous
    import matplotlib.pyplot as plt
    hospital_prior_plot_exogenous(f=quick_precision_ema_ensemble,k=k)
    plt.show()
    pass



from timemachines.skaters.localskaters import LOCAL_SKATERS, FAST_LOCAL_SKATERS
from timemachines.skaters.networkedskaters import NETWORKED_SKATERS, FAST_NETWORKED_SKATERS

SKATERS = NETWORKED_SKATERS + LOCAL_SKATERS
FAST_SKATERS = FAST_LOCAL_SKATERS + FAST_NETWORKED_SKATERS


def skater_from_name(name):
    valid = [f for f in SKATERS if f.__name__ == name]
    return valid[0] if len(valid)==1 else None




from timemachines.skaters.proph.prophskaterssingular import PROPHET_SKATERS_SINGULAR, PROPHET_R2_SKATERS
from timemachines.skaters.proph.prophskaterscomposed import PROPHET_SKATERS_COMPOSED


PROPHET_SKATERS = PROPHET_SKATERS_SINGULAR + PROPHET_SKATERS_COMPOSED





from timemachines.skaters.rvr.rvrinclusion import using_river
from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE, wrap
from typing import Any
from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous

if using_river:
    from timemachines.skaters.rvr.rvrsarimax import rvr_sarimax_factory

    def rvr_p1_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                             t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return rvr_sarimax_factory(y=y,s=s,k=k,a=a,t=t,e=e,p=1,d=0,q=0)


    def rvr_p2_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                     t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return rvr_sarimax_factory(y=y, s=s, k=k, a=a, t=t, e=e, p=2, d=0, q=0)


    def rvr_p3_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                     t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return rvr_sarimax_factory(y=y, s=s, k=k, a=a, t=t, e=e, p=3, d=0, q=0)


    def rvr_p5_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                     t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return rvr_sarimax_factory(y=y, s=s, k=k, a=a, t=t, e=e, p=5, d=0, q=0)

    def rvr_p8_d0_q0(y: Y_TYPE, s: dict, k: int, a: A_TYPE = None,
                     t: T_TYPE = None, e: E_TYPE = None) -> ([float], Any, Any):
        return rvr_sarimax_factory(y=y, s=s, k=k, a=a, t=t, e=e, p=5, d=0, q=0)


    RIVER_CONSTANT_SKATERS = [rvr_p1_d0_q0, rvr_p2_d0_q0, rvr_p3_d0_q0, rvr_p3_d0_q0, rvr_p5_d0_q0, rvr_p8_d0_q0]
else:
    RIVER_CONSTANT_SKATERS = []

if __name__=='__main__':
    prior_plot_exogenous(f=rvr_p1_d0_q0,k=5,n=50,n_plot=50)


from timemachines.skaters.drts.dartsinclusion import using_darts

if using_darts:
    import numpy as np
    import pandas as pd
    from typing import List
    from darts import TimeSeries
    from darts.models import FFT,Prophet,ARIMA,AutoARIMA,ExponentialSmoothing,Theta,FourTheta,TransformerModel

    def darts_FFT_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls Darts' FFT forecasting model, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        y0s.extend([0]*k)
        idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
        df = pd.DataFrame(y0s, index=idx).reset_index().rename(columns={'index':'ts',0:'y'})
        df.head()

        series = TimeSeries.from_dataframe(df, 'ts', 'y')
        train, val = series[:-k], series[-k:]

        model = FFT()
        model.fit(train)
        prediction = model.predict(len(val))

        x = prediction.pd_dataframe()['y'].values.tolist()
        x_std = [1]*k

        return x, x_std
        
    def darts_Prophet_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls Darts' Prophet forecasting model, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        y0s.extend([0]*k)
        idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
        df = pd.DataFrame(y0s, index=idx).reset_index().rename(columns={'index':'ts',0:'y'})
        df.head()

        series = TimeSeries.from_dataframe(df, 'ts', 'y')
        train, val = series[:-k], series[-k:]

        model = Prophet()
        model.fit(train)
        prediction = model.predict(len(val))

        x = prediction.pd_dataframe()['y'].values.tolist()
        x_std = [1]*k

        return x, x_std
        
    def darts_ARIMA_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls Darts' ARIMA forecasting model, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        y0s.extend([0]*k)
        idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
        df = pd.DataFrame(y0s, index=idx).reset_index().rename(columns={'index':'ts',0:'y'})
        df.head()

        series = TimeSeries.from_dataframe(df, 'ts', 'y')
        train, val = series[:-k], series[-k:]

        model = ARIMA()
        model.fit(train)
        prediction = model.predict(len(val))

        x = prediction.pd_dataframe()['y'].values.tolist()
        x_std = [1]*k

        return x, x_std
        
    def darts_AutoARIMA_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls Darts' AutoARIMA forecasting model, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        y0s.extend([0]*k)
        idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
        df = pd.DataFrame(y0s, index=idx).reset_index().rename(columns={'index':'ts',0:'y'})
        df.head()

        series = TimeSeries.from_dataframe(df, 'ts', 'y')
        train, val = series[:-k], series[-k:]

        model = AutoARIMA()
        model.fit(train)
        prediction = model.predict(len(val))

        x = prediction.pd_dataframe()['y'].values.tolist()
        x_std = [1]*k

        return x, x_std
        
    def darts_ExponentialSmoothing_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls Darts' ExponentialSmoothing forecasting model, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        y0s.extend([0]*k)
        idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
        df = pd.DataFrame(y0s, index=idx).reset_index().rename(columns={'index':'ts',0:'y'})
        df.head()

        series = TimeSeries.from_dataframe(df, 'ts', 'y')
        train, val = series[:-k], series[-k:]

        model = ExponentialSmoothing()
        model.fit(train)
        prediction = model.predict(len(val))

        x = prediction.pd_dataframe()['y'].values.tolist()
        x_std = [1]*k

        return x, x_std
        
    def darts_Theta_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls Darts' Theta forecasting model, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        y0s.extend([0]*k)
        idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
        df = pd.DataFrame(y0s, index=idx).reset_index().rename(columns={'index':'ts',0:'y'})
        df.head()

        series = TimeSeries.from_dataframe(df, 'ts', 'y')
        train, val = series[:-k], series[-k:]

        model = Theta()
        model.fit(train)
        prediction = model.predict(len(val))

        x = prediction.pd_dataframe()['y'].values.tolist()
        x_std = [1]*k

        return x, x_std
        
    def darts_FourTheta_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls Darts' FourTheta forecasting model, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        y0s.extend([0]*k)
        idx = pd.date_range(end='01/01/2000', periods=len(y0s), freq="H")
        df = pd.DataFrame(y0s, index=idx).reset_index().rename(columns={'index':'ts',0:'y'})
        df.head()

        series = TimeSeries.from_dataframe(df, 'ts', 'y')
        train, val = series[:-k], series[-k:]

        model = FourTheta()
        model.fit(train)
        prediction = model.predict(len(val))

        x = prediction.pd_dataframe()['y'].values.tolist()
        x_std = [1]*k

        return x, x_std
        

if __name__=='__main__':
    assert using_darts, 'pip install \'u8darts[all]\''
    import numpy as np
    import time
    st = time.time()
    y = list(np.cumsum(np.random.randn(400)))
    x, x_std = darts_FFT_iskater(y=y, k=5)
    print(x)
    print(time.time()-st)


from timemachines.skaters.pmd.pmdinclusion import using_pmd, pm
if using_pmd:
    from timemachines.skaters.pmd.pmdskaterfactory import pmd_skater_factory
    from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE,T_TYPE,E_TYPE, wrap
    from timemachines.skaters.simple.movingaverage import empirical_last_value
    from timemachines.skatertools.visualization.priorplot import prior_plot, prior_plot_exogenous


    def pmd_exogenous(y:Y_TYPE, s:dict, k:int=1, a:A_TYPE=None, t:T_TYPE=None, e:E_TYPE=None):
        """ Predict using auto_arima, with both simultaneously observed and known in advance variables
            This skater has no hyper-parameters

            y: Y_TYPE    scalar or list where y[1:] are interpreted as contemporaneously observed exogenous variables
            s:           state
            k:           Number of steps ahead to predict
            a:           (optional) scalar or list of variables known k-steps in advance.
                          (IMPORTANT: If supplying 'a', provide the known variable k steps ahead, not the contemporaneous one !).
            t:           (optional) Time of observation.
            e:           (optional) Maximum computation time (supply e>60 to give hint to do fitting)

            :returns: x [float] , s', scale [float]
        """
        return pmd_skater_factory(y=y, s=s, k=k, a=a, t=t, e=e, method='auto')


    def pmd_known(y:Y_TYPE, s:dict, k:int=1, a:A_TYPE=None, t:T_TYPE=None, e:E_TYPE=None):
        """ Uses known-in-advance but not y[1:] """
        y0 = [wrap(y)[0]]
        return pmd_skater_factory(y=y0, s=s, k=k, a=a, t=t, e=e, method='auto')


    def pmd_univariate(y:Y_TYPE, s:dict, k:int=1, a:A_TYPE=None, t:T_TYPE=None, e:E_TYPE=None):
        """ Uses only y[0] and ignores y[1:] and a[:] """
        y0 = [wrap(y)[0]]
        return pmd_skater_factory(y=y0, s=s, k=k, a=None, t=t, e=e, method='auto')


    def pmd_exog_compare(f,k=1):
        from timemachines.skatertools.evaluation.evaluators import evaluate_mean_absolute_error
        from timemachines.skatertools.evaluation.evaluators import hospital_with_exog
        y, a = hospital_with_exog(k=k)
        y0 = [ yi[0] for yi in y ]

        r = 0.1 # Doesn't matter?
        err1 = evaluate_mean_absolute_error(f=f, k=k, y=y0, r=r, n_burn=250)
        err2 = evaluate_mean_absolute_error(f=f, k=k, y=y, r=r, n_burn=250)
        err3 = evaluate_mean_absolute_error(f=f, k=k, y=y, r=r, a=a, n_burn=250)
        errlv = evaluate_mean_absolute_error(f=empirical_last_value, k=k, y=y, r=r, a=a, n_burn=250)


        print('----------------')
        print("Error w/o exogenous   = "+str(err1))
        print("Error w   exogenous   = "+str(err2))
        print("Error w   exo + known = "+str(err3))
        print("Error last val cache  = " + str(errlv))


if __name__ == '__main__':
    assert using_pmd,'pip install pmdarima'
    f = pmd_exogenous
    if True:
        prior_plot_exogenous(f=f, k=1, n=200)
    if True:
        prior_plot(f=f,k=1,n=200)



from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, T_TYPE, E_TYPE
from typing import Any
from timemachines.skatertools.recommendations.suggestions import top_rated_models
from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory
from timemachines.skatertools.utilities.internet import connected_to_internet


def elo_ensemble_factory(y: Y_TYPE, s: dict, k: int = 1, a:A_TYPE = None, t:T_TYPE = None, e:E_TYPE = None,r=None,
                         category='univariate',
                         max_seconds=1.0,
                         min_count=10,
                         require_passing=True,
                         n=5)->([float], Any, Any):
    """
        Ensemble of top performing skaters.

        r - determines variance weighting
        See top_rated_models for other argument meanings
    """
    assert r is not None, 'need the precision coef'
    if s.get('top_rated_skaters') is None:
        assert connected_to_internet(), 'cannot get best skaters without internet connection'
        top_rated_skaters = top_rated_models(k=k,n=n,category=category,max_seconds=max_seconds,min_count=min_count,
                                         require_passing=require_passing, ignore_elo=True)
        if len(top_rated_skaters)<=1:
            raise Exception('Could not find enough rated models')
        s['top_rated_skaters'] = top_rated_skaters


    return precision_weighted_ensemble_factory(fs=s['top_rated_skaters'], y=y, s=s, k=k, a=a, t=t, e=e, r=r)





from timemachines.skaters.dlm.dlminclusion import using_dlm, dlm, trend, seasonality, autoReg
from timemachines.skatertools.utilities.conventions import dimension, to_int_log_space, split_exogenous, Y_TYPE, R_TYPE
from timemachines.skatertools.visualization.priorplot import prior_plot_exogenous
import math
import numpy as np

if using_dlm:

    from timemachines.skaters.dlm.dlmhacks import fixedAutoReg

    ###################################################
    #                                                 #
    #      Univariate skater based on pydlm           #
    #                                                 #
    ###################################################


    def dlm_univariate_a(y, s: dict, k: int, a=None, t=None, e=None):
        ra = 0.25
        return dlm_univariate_r3(y=y, s=s, k=k, r=ra)


    def dlm_univariate_b(y, s: dict, k: int, a=None, t=None, e=None):
        ra = 0.75
        return dlm_univariate_r3(y=y, s=s, k=k, r=ra)


    def dlm_univariate_r3(y, s: dict, k: int, a=None, t=None, e=None, r=None):
        """ Univariate filter

                - Uses the discounting method of H/W so, doesn't need to be fit as often
                - Discount factors are periodically tuned
                - The hyper-parameter controls 'auto_degree', 'trend_degree',  'period'

            :returns: x, x_std, s

            This guy will take a hint if e>10 but regardless, will occasionally perform a tuning
            of parameters.

        """
        assert r is not None, 'Requires hyper-parameter (interpreted in dimension 3) '
        if not s:
            s = dict()
            s = dlm_set_univariate_params(s=s, r=r)
            s['dim'] = dimension(y)
            s['n_obs'] = 0
            s['model'] = dlm([], printInfo=False) + trend(s['trend_degree'], s['discount']) + seasonality(s['period'],
                                                                                                          s['discount'])
            s['model'] = s['model'] + fixedAutoReg(degree=s['auto_degree'], name='ar', w=1.0)

        if y is not None:
            s['n_obs'] += 1
            assert isinstance(y, float) or len(y) == s['dim'], ' Cannot change dimension of input in flight '
            y0, exog = split_exogenous(y=y)
            y0_passed_in = None if np.isnan(y0) else y0  # pydlm uses None for missing values
            s['model'].append([y0_passed_in])
            num_obs = len(s['model'].data) if s.get('model') else 0
            if (num_obs % s['n_fit'] == s['n_fit'] - 1 ) or (e is not None and e>10):
                # Perform periodic tuning of discount factors
                _, _, s = dlm_univariate_r3(y=None, s=s, k=k, a=a, t=t, e=1000, r=r)
            s['model'].fitForwardFilter()
            return _dlm_prediction_helper(s=s, k=k, y=y)

        if y is None and e > 0:
            # Not sure I like this style, versus a separate fit function for clarity
            s['model'].tune()  # Tunes discount factors
            s['model'].fit()
            return None, None, s


    def _dlm_prediction_helper(s, k:int, y:Y_TYPE):
        """ Calls down to predictN, if we are passed the warm-up stage """
        num_obs = len(s['model'].data) if s.get('model') else 0
        if num_obs < s['n_warm']:
            y0, _ = split_exogenous(y)
            return [y0]*k, [abs(y0)]*k, s
        else:
            x_mean, x_var = s['model'].predictN(N=k)
            x = list(x_mean)
            x_std = [ math.sqrt(v) for v in x_var]
            return x, x_std, s



    def dlm_set_univariate_params(s, r:R_TYPE):
        period_choices = [3,5,7,10,12,16,24,32]
        s['auto_degree'], s['trend_degree'], period_choice = to_int_log_space(r,bounds=[
            (0.5,3.2),    # auto-reg degree. 1 means linear trend, 2 quadratic
            (0.1,3.4),    # seasonality degree
            (0,len(period_choices)) # seasonality period
        ])
        s['discount'] = 1 - ( r % 0.05 ) # Shouldn't matter too much as these get tuned eventually
        s['period'] = period_choices[period_choice]
        s['n_warm'] = 100
        s['n_fit']  = 500  # How often to tune discounts subsequently
        return s


if __name__ == '__main__':
    assert using_dlm, 'pip install pydlm'
    import matplotlib.pyplot as plt
    err = prior_plot_exogenous(f=dlm_univariate_r3, k=2, n=500, r=np.random.rand())
    plt.figure()
    print('done')
    pass



from timemachines.skatertools.utilities.conventions import Y_TYPE, A_TYPE, R_TYPE, E_TYPE, T_TYPE
from timemachines.skaters.tsa.tsaconstant import TSA_D0_SKATERS
from timemachines.skatertools.ensembling.ensemblefactory import precision_weighted_ensemble_factory
from timemachines.skatertools.utilities.conventions import wrap


def hypocratic_tsa_d0_factory(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None, r:R_TYPE=None):
    """
         r :  moving average parameter  (e.g. 0.75 is fast, 0.95 is slow)
    """
    y0 = wrap(y)[0]
    assert r is not None
    x, x_std, s = precision_weighted_ensemble_factory(fs=TSA_D0_SKATERS, y=y0, s=s, k=k, a=a, t=t, e=e, r=r)

    def hypocratic(x:float, x_std:float, confidence=0.5):
        """ Shrink residual prediction towards zero """
        # Remains to be seen if this is sensible for TSA residuals
        import math
        if abs(x_std)<1e-6 or abs(x)<1e-3*x_std:
            return 0.0
        else:
            return x*math.tanh(confidence*abs(x)/(3*x_std))

    x_resid = [ hypocratic(xi,x_std) for xi,x_std in zip(x,x_std) ]
    return x_resid, x_std, s


def tsa_slowly_hypocratic_d0_ensemble(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None):
    return hypocratic_tsa_d0_factory(y=y,s=s,k=k,a=a,t=t,e=e,r=0.95)


def tsa_quickly_hypocratic_d0_ensemble(y :Y_TYPE, s, k:int, a:A_TYPE =None, t:T_TYPE =None, e:E_TYPE =None):
    return hypocratic_tsa_d0_factory(y=y,s=s,k=k,a=a,t=t,e=e,r=0.75)


TSA_DO_HYPOCRATIC_SKATERS = [tsa_slowly_hypocratic_d0_ensemble, tsa_quickly_hypocratic_d0_ensemble]


from timemachines.skaters.gk.greykiteinclusion import using_greykite

if using_greykite:
    import numpy as np
    import pandas as pd
    from typing import List
    import datetime
    from timemachines.skatertools.utilities.suppression import no_stdout_stderr
    from greykite.framework.templates.autogen.forecast_config import ForecastConfig
    from greykite.framework.templates.autogen.forecast_config import MetadataParam
    from greykite.framework.templates.forecaster import Forecaster 
    from greykite.framework.templates.model_templates import ModelTemplateEnum
    from greykite.framework.utils.result_summary import summarize_grid_search_results

    def linkedin_greykite_iskater(y: [[float]], k: int, a: List = None, t: List = None, e=None, deseasonalize=False):
        """
            Calls Linkedin Greykite's forecasting model, but ignores t if supplied.
        """
        if a:
            assert len(a) == len(y) + k
        if np.isscalar(y[0]):
            y0s = [ yt for yt in y]
        else:
            y0s = [ yt[0] for yt in y ]

        idx = pd.date_range("1970-01-01", periods=len(y0s), freq="H")
        df = pd.DataFrame(y0s, index=idx).reset_index().rename(columns={'index':'ts',0:'y'})

        metadata = MetadataParam(
            time_col="ts",
            value_col="y",
            freq="H"
        )

        forecaster = Forecaster()
        result = forecaster.run_forecast_config(
             df=df,
             config=ForecastConfig(
                 model_template=ModelTemplateEnum.SILVERKITE.name,
                 forecast_horizon=k,
                 coverage=0.68,
                 metadata_param=metadata
             )
         )

        forecast = result.forecast
        x = list(forecast.df[-k:].forecast)
        x_std = list((forecast.df.forecast_upper - forecast.df.forecast_lower)/2)[-k:]
        return x, x_std

if __name__=='__main__':
    assert using_greykite, 'pip install greykite'
    import numpy as np
    import time
    st = time.time()
    y = list(np.cumsum(np.random.randn(400)))
    x, x_std = linkedin_greykite_iskater(y=y, k=5)
    print(x)
    print(time.time()-st)